from ..conversable_agent import ConversableAgent
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union, Generator
from ....utils.client import ByzerLLM
from byzerllm.utils.retrieval import ByzerRetrieval
from ..agent import Agent
import ray
from ray.util.client.common import ClientActorHandle, ClientObjectRef
from .. import get_agent_name,run_agent_func,ChatResponse
from byzerllm.apps.agent.extensions.simple_retrieval_client import SimpleRetrievalClient
import uuid
import json
from byzerllm.apps.llama_index import get_service_context,get_storage_context
from llama_index import VectorStoreIndex


try:
    from termcolor import colored
except ImportError:

    def colored(x, *args, **kwargs):
        return x

import jieba     
import pydantic

class AgentData(pydantic.BaseModel):        
    namespace:str = pydantic.Field(...,description="用户提及的命名空间名字,如果没有提及，则设置为 default")        


class LlamaIndexRetrievalAgent(ConversableAgent): 
    PROMPT_DEFAULT = """You're a retrieve augmented chatbot. """    
    DEFAULT_SYSTEM_MESSAGE = PROMPT_DEFAULT
    
    def __init__(
        self,
        name: str,
        llm: ByzerLLM,        
        retrieval: ByzerRetrieval,        
        chat_name:str,
        owner:str,                
        update_context_retry: int = 3,        
        system_message: Optional[str] = DEFAULT_SYSTEM_MESSAGE,        
        is_termination_msg: Optional[Callable[[Dict], bool]] = None,
        max_consecutive_auto_reply: Optional[int] = None,
        human_input_mode: Optional[str] = "NEVER",
        code_execution_config: Optional[Union[Dict, bool]] = False,
        **kwargs,
    ):       
        super().__init__(
            name,
            llm,retrieval,
            system_message,
            is_termination_msg,
            max_consecutive_auto_reply,
            human_input_mode,
            code_execution_config=code_execution_config,            
            **kwargs,
        )
        self.chat_name = chat_name
        self.owner = owner        
        self.update_context_retry = update_context_retry


        self._reply_func_list = []
        # self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.generate_llm_reply)   
        self.register_reply([Agent, ClientActorHandle,str], LlamaIndexRetrievalAgent.generate_retrieval_based_reply) 
        self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.check_termination_and_human_reply) 
        self.llm = llm
        self.retrieval = retrieval                                                         

    def generate_retrieval_based_reply(
        self,
        raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Union[ClientActorHandle,Agent,str]] = None,
        config: Optional[Any] = None,
    ) -> Tuple[bool, Union[str, Dict, None,ChatResponse]]:  
        
        if messages is None:
            messages = self._messages[get_agent_name(sender)]
                
        new_message = messages[-1]
        content = new_message["content"]
        
        # @self.llm.response()
        # def extract_data(s:str)->AgentData:
        #     pass

        # agent_data = extract_data(content)
        agent_data = AgentData(namespace="default")
        if not agent_data.namespace:
            agent_data.namespace = "default" 

        print(f"{agent_data}",flush=True)    

        service_context = get_service_context(self.llm)
        storage_context = get_storage_context(self.llm,self.retrieval,chunk_collection=agent_data.namespace,namespace=agent_data.namespace)                

        index = VectorStoreIndex.from_vector_store(vector_store = storage_context.vector_store,service_context=service_context)
        query_engine = index.as_query_engine(streaming=True)        
        id = str(uuid.uuid4())        
        streaming_response = query_engine.query(content)
        
        def gen(): 
            t = ""           
            for response in streaming_response.response_gen:
                t += response
                yield (t,None)

        return self.stream_reply(gen(),contexts=[])  
        
        
                
        
                    


from byzerllm.utils.retrieval import ByzerRetrieval
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from byzerllm.utils.client import ByzerLLM
from byzerllm.utils import generate_str_md5
from byzerllm.utils.retrieval import ByzerRetrieval
import time
from byzerllm.utils.client import LLMHistoryItem, LLMRequest
from byzerllm.utils.retrieval import TableSettings, SearchQuery
import uuid
import json
from langchain_core.prompts import PromptTemplate
from langchain.text_splitter import RecursiveCharacterTextSplitter

try:
    from termcolor import colored
except ImportError:

    def colored(x, *args, **kwargs):
        return x


try:
    import jieba
except ImportError:
    colored(
        "Warning: jieba module is not installed. You may need to install it by running `pip install jieba`"
    )
    jieba = None


class SimpleRetrieval:
    def __init__(
        self,
        llm: ByzerLLM,
        retrieval: ByzerRetrieval,
        chunk_collection: Optional[str] = "default",
        retrieval_cluster: str = "byzerai_store",
        retrieval_db: str = "byzerai_store",
        max_output_length=10000,
    ):
        self.chunk_collection = chunk_collection
        self.retrieval_cluster = retrieval_cluster
        self.retrieval_db = retrieval_db
        self.max_output_length = max_output_length
        self.llm = llm
        self.retrieval = retrieval
        if self.llm.default_emb_model_name is None:
            raise Exception(
                f"""
emb model does not exist. Try to use `llm.setup_default_emb_model_name` to set the default emb model name.
"""
            )

        # create the retrieval database/table if not exists
        if self.retrieval and not self.retrieval.check_table_exists(
            self.retrieval_cluster, self.retrieval_db, "text_content"
        ):
            self.retrieval.create_table(
                self.retrieval_cluster,
                tableSettings=TableSettings(
                    database=self.retrieval_db,
                    table="text_content",
                    schema="""st(
field(_id,string),
field(doc_id,string),
field(file_path,string),
field(owner,string),
field(content,string,analyze),
field(raw_content,string,no_index),
field(metadata,string,analyze),
field(json_data,string,no_index),
field(collection,string),
field(created_time,long,sort)
)""",
                    location=f"/tmp/{self.retrieval_cluster}",
                    num_shards=1,
                ),
            )

            self.retrieval.create_table(
                self.retrieval_cluster,
                tableSettings=TableSettings(
                    database=self.retrieval_db,
                    table="text_content_chunk",
                    schema="""st(
field(_id,string),
field(doc_id,string),
field(file_path,string),
field(owner,string),
field(chunk,string,analyze),
field(raw_chunk,string,no_index),
field(chunk_vector,array(float)),
field(metadata,string,analyze),
field(json_data,string,no_index),
field(chunk_collection,string),
field(created_time,long,sort)
)""",
                    location=f"/tmp/{self.retrieval_cluster}",
                    num_shards=1,
                ),
            )

    def save_doc(
        self,
        data: List[Dict[str, Any]],
        owner: Optional[str] = None,
    ):

        if not self.retrieval:
            raise Exception("retrieval is not setup")

        owner = owner or "default"

        result = []
        for item in data:

            doc_id = item["doc_id"]
            json_data_obj = json.loads(item["json_data"])
            collection = item["collection"]
            _id = f"{collection}/{doc_id}"

            file_path = json_data_obj.get("metadata", {}).get("file_path", "")
            result.append(
                {
                    "_id": _id,
                    "doc_id": doc_id,
                    "file_path": file_path,
                    "json_data": item["json_data"],
                    "metadata": self.search_tokenize(item["json_data"]),
                    "collection": collection,
                    "content": self.search_tokenize(item["content"]),
                    "raw_content": item["content"],
                    "owner": owner,
                    "created_time": int(time.time() * 1000),
                }
            )

        self.retrieval.build_from_dicts(
            self.retrieval_cluster, self.retrieval_db, "text_content", result
        )

    def _owner_filter(self, owner: str):
        return {"field": "owner", "value": owner}

    def search_content_chunks(
        self,
        owner: str,
        query_str: Optional[str] = None,
        query_embedding: Optional[List[float]] = None,
        doc_ids: Optional[List[str]] = None,
        limit: int = 4,
        return_json: bool = True,
    ):
        keyword = None
        fields = []
        vector = []
        vectorField = None
        filters = {
            "and": [{"field": "chunk_collection", "value": self.chunk_collection}]
        }

        if query_str is not None:
            keyword = self.search_tokenize(query_str)
            fields = ["chunk"]

        if query_embedding is not None:
            vector = query_embedding
            vectorField = "chunk_vector"

        if doc_ids:
            filters["and"].append(
                {"or": [{"field": "doc_id", "value": x} for x in doc_ids]}
            )

        query = SearchQuery(
            self.retrieval_db,
            "text_content_chunk",
            filters=filters,
            keyword=keyword,
            fields=fields,
            vector=vector,
            vectorField=vectorField,
            limit=limit,
        )

        docs = self.retrieval.search(self.retrieval_cluster, [query])

        if return_json:
            context = json.dumps(
                [{"content": x["raw_chunk"]} for x in docs],
                ensure_ascii=False,
                indent=4,
            )
            return context
        else:
            return docs

    def search_content_by_filename(self, filename: str, collection: str):
        filters = [{"field": "chunk_collection", "value": collection}]
        docs = self.retrieval.search(
            self.retrieval_cluster,
            [
                SearchQuery(
                    self.retrieval_db,
                    "text_content_chunk",
                    filters={"and": filters},
                    keyword=self.search_tokenize(filename),
                    fields=["metadata"],
                    vector=[],
                    vectorField=None,
                    limit=10000,
                )
            ],
        )
        return docs

    def search_content(
        self,
        q: str,
        owner: str,
        url: str,
        auth_tag: str = None,
        limit: int = 4,
        return_json: bool = True,
    ):
        filters = [self._owner_filter(owner)]

        if auth_tag:
            filters.append(
                {"field": "auth_tag", "value": self.search_tokenize(auth_tag)}
            )

        if url:
            filters.append({"field": "url", "value": url})

        if q:
            keyword = self.search_tokenize(q)
            vector = self.emb(q)
            vectorField = "content_vector"
            fields = ["content"]
        else:
            keyword = None
            vector = []
            vectorField = None
            fields = []

        docs = self.retrieval.search(
            self.retrieval_cluster,
            [
                SearchQuery(
                    self.retrieval_db,
                    "text_content",
                    filters={"and": filters},
                    keyword=keyword,
                    fields=fields,
                    vector=vector,
                    vectorField=vectorField,
                    limit=limit,
                )
            ],
        )

        if return_json:
            context = json.dumps(
                [{"content": x["raw_content"]} for x in docs],
                ensure_ascii=False,
                indent=4,
            )
            return context
        else:
            return docs

    def get_chunk_by_id(self, chunk_id: str):
        filters = {"and": [{"field": "_id", "value": chunk_id}]}
        docs = self.retrieval.filter(
            self.retrieval_cluster,
            [
                SearchQuery(
                    self.retrieval_db,
                    "text_content_chunk",
                    filters=filters,
                    keyword=None,
                    fields=[],
                    vector=[],
                    vectorField=None,
                    limit=1,
                )
            ],
        )
        return docs

    def get_chunks_by_docid(self, doc_id: str):
        docs = self.retrieval.filter(
            self.retrieval_cluster,
            [
                SearchQuery(
                    self.retrieval_db,
                    "text_content_chunk",
                    filters={"and": [{"field": "doc_id", "value": doc_id}]},
                    keyword=None,
                    fields=[],
                    vector=[],
                    vectorField=None,
                    limit=100000,
                )
            ],
        )
        return docs

    def delete_chunk_by_ids(self, chunk_ids: List[str]):
        self.retrieval.delete_by_ids(
            self.retrieval_cluster, self.retrieval_db, "text_content_chunk", chunk_ids
        )
        self.retrieval.commit(self.retrieval_cluster, self.retrieval_db)

    def save_chunks(self, chunks: List[Dict[str, Any]]):
        text_content_chunks = []
        for chunk in chunks:
            chunk_id = chunk["chunk_id"]
            ref_doc_id = chunk["ref_doc_id"]
            owner = chunk.get("owner", "default")
            chunk_content = chunk["chunk_content"]
            chunk_embedding = chunk["chunk_embedding"]
            metadata = chunk.get("metadata", {})
            _id = f"{self.chunk_collection}/{chunk_id}"
            file_path = metadata.get("file_path", "")

            text_content_chunks.append(
                {
                    "_id": _id,
                    "doc_id": ref_doc_id or "",
                    "file_path": file_path,
                    "owner": owner or "default",
                    "chunk": self.search_tokenize(chunk_content),
                    "raw_chunk": chunk_content,
                    "chunk_vector": chunk_embedding,
                    "chunk_collection": self.chunk_collection,
                    "metadata": self.search_tokenize(
                        json.dumps(metadata, ensure_ascii=False)
                    ),
                    "json_data": json.dumps(metadata, ensure_ascii=False),
                    "created_time": int(time.time() * 1000),
                }
            )

        self.retrieval.build_from_dicts(
            self.retrieval_cluster,
            self.retrieval_db,
            "text_content_chunk",
            text_content_chunks,
        )

    def get_doc(self, doc_id: str, collection: str):
        filters = {"and": [{"field": "_id", "value": f"{collection}/{doc_id}"}]}
        docs = self.retrieval.filter(
            self.retrieval_cluster,
            [
                SearchQuery(
                    self.retrieval_db,
                    "text_content",
                    filters=filters,
                    keyword=None,
                    fields=[],
                    vector=[],
                    vectorField=None,
                    limit=1,
                )
            ],
        )
        return docs[0] if docs else None

    @DeprecationWarning
    def delete_doc(self, doc_ids: List[str], collection: str):
        ids = []
        for doc_id in doc_ids:
            ids.append(f"{collection}/{doc_id}")
        self.retrieval.delete_by_ids(
            self.retrieval_cluster, self.retrieval_db, "text_content", [ids]
        )

    def truncate_table(self):
        self.retrieval.truncate(
            self.retrieval_cluster, self.retrieval_db, "text_content"
        )
        self.retrieval.truncate(
            self.retrieval_cluster, self.retrieval_db, "text_content_chunk"
        )
        self.commit_doc()
        self.commit_chunk()

    def delete_doc_and_chunks_by_filename(self, collection: str, file_name: str):
        doc = self.get_doc(f"{collection}/ref_doc_info", file_name)

        if doc:
            node_ids = json.loads(doc["metadata"])["node_ids"]
            for node_id in node_ids:
                id = f"{collection}/data/{node_id}"
                self.retrieval.delete_by_ids(
                    self.retrieval_cluster, self.retrieval_db, "text_content", [id]
                )

                id = f"{collection}/metadata/{node_id}"
                self.retrieval.delete_by_ids(
                    self.retrieval_cluster, self.retrieval_db, "text_content", [id]
                )

                ## default/index/ can not be deleted

            ##  cleanup the chunk collection
            self.retrieval.delete_by_filter(
                self.retrieval_cluster,
                self.retrieval_db,
                "text_content_chunk",
                {"file_path": file_name},
            )

    def delete_from_doc_collection(self, collection: str):
        for suffix in ["index", "data", "ref_doc_info", "metadata"]:
            self.retrieval.delete_by_filter(
                self.retrieval_cluster,
                self.retrieval_db,
                "text_content",
                {"collection": f"{collection}/{suffix}"},
            )

    def delete_from_chunk_collection(self, collection: str):
        self.retrieval.delete_by_filter(
            self.retrieval_cluster,
            self.retrieval_db,
            "text_content_chunk",
            {"chunk_collection": collection},
        )

    def commit_doc(self):
        self.retrieval.commit(self.retrieval_cluster, self.retrieval_db, "text_content")

    def commit_chunk(self):
        self.retrieval.commit(
            self.retrieval_cluster, self.retrieval_db, "text_content_chunk"
        )

    def emb(self, s: str):
        return self.llm.emb(self.llm.default_emb_model_name, LLMRequest(instruction=s))[
            0
        ].output

    def split_text_into_chunks(self, s: str):
        # self.llm.apply_sql_func(
        #     '''select llm_split(value,array(",","。","\n"),1600) as value ''',[{"value":content}],
        #     url=self.byzer_engine_url
        #     )["value"]
        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=1600, chunk_overlap=200
        )
        split_docs = text_splitter.split_documents([Document(page_content=s)])
        return [s.page_content for s in split_docs]

    def search_tokenize(self, s: str):
        seg_list = jieba.cut(s, cut_all=False)
        # return self.llm.apply_sql_func("select mkString(' ',parse(value)) as value",[
        # {"value":s}],url=self.byzer_engine_url)["value"]
        return " ".join(seg_list)



import gc
import os
from typing import *

import torch
from diffusers import AutoencoderKL, UNet2DConditionModel
from diffusers.pipelines.stable_diffusion import convert_from_ckpt
from transformers import CLIPTextModel
from byzerllm.stable_diffusion.config import stableDiffusionConfig

from byzerllm.stable_diffusion.shared import (
    hf_diffusers_cache_dir,
    hf_transformers_cache_dir,
)


def convert_checkpoint_to_pipe(model_id: str):
    if stableDiffusionConfig.get_checkpoint():
        if os.path.exists(model_id) and os.path.exists(model_id):
            return convert_from_ckpt.download_from_original_stable_diffusion_ckpt(
                model_id,
                from_safetensors=model_id.endswith(".safetensors"),
                load_safety_checker=False,
            )
    else:
        raise Exception(
            f"No {model_id} found.In checkpoint mode, model_dir must be a file.Please set the checkpoint path in the model_dir config."
        )


def load_unet(
    model_id: str, device: Optional[torch.device] = None
) -> UNet2DConditionModel:
    temporary_pipe = convert_checkpoint_to_pipe(model_id)
    if temporary_pipe is not None:
        unet = temporary_pipe.unet
        del temporary_pipe
        gc.collect()
        torch.cuda.empty_cache()
    else:
        unet = UNet2DConditionModel.from_pretrained(
            model_id, subfolder="unet", cache_dir=hf_diffusers_cache_dir()
        )
    unet = unet.to(device=device)
    return unet


def load_text_encoder(model_id: str, device: Optional[torch.device] = None):
    temporary_pipe = convert_checkpoint_to_pipe(model_id)
    if temporary_pipe is not None:
        text_encoder = temporary_pipe.text_encoder
        del temporary_pipe
        gc.collect()
        torch.cuda.empty_cache()
    else:
        text_encoder = CLIPTextModel.from_pretrained(
            model_id, subfolder="text_encoder", cache_dir=hf_transformers_cache_dir()
        )
    text_encoder = text_encoder.to(device=device)
    return text_encoder


def load_vae_decoder(model_id: str, device: Optional[torch.device] = None):
    temporary_pipe = convert_checkpoint_to_pipe(model_id)
    if temporary_pipe is not None:
        vae = temporary_pipe.vae
        del temporary_pipe
        gc.collect()
        torch.cuda.empty_cache()
    else:
        vae = AutoencoderKL.from_pretrained(
            model_id, subfolder="vae", cache_dir=hf_diffusers_cache_dir()
        )

    vae.forward = vae.decode
    vae = vae.to(device=device)
    return vae


def load_vae_encoder(model_id: str, device: Optional[torch.device] = None):
    temporary_pipe = convert_checkpoint_to_pipe(model_id)
    if temporary_pipe is not None:
        vae = temporary_pipe.vae
        del temporary_pipe
        gc.collect()
        torch.cuda.empty_cache()
    else:
        vae = AutoencoderKL.from_pretrained(
            model_id, subfolder="vae", cache_dir=hf_diffusers_cache_dir()
        )

    def encoder_forward(x):
        return vae.encode(x).latent_dist.sample()

    vae.forward = encoder_forward
    vae = vae.to(device=device)
    return vae



from typing import Any, List, Dict
from ray.util.client.common import ClientObjectRef
from pyjava.api.mlsql import RayContext
from pyjava.storage import streaming_tar
import os
import inspect
import functools
import json
from typing import Type, Any

from typing import Dict, Generator, Optional
from dataclasses import dataclass
from byzerllm.utils import (
    print_flush,
    format_prompt,
    format_prompt_jinja2,
    format_str_jinja2,
)
from .store import transfer_from_ob


@dataclass
class BlockRow:
    start: int
    offset: int
    value: bytes


def restore_model(conf: Dict[str, str], target_dir: str):
    model_servers = RayContext.parse_servers(conf["modelServers"])
    model_binary = RayContext.collect_from(model_servers)
    streaming_tar.save_rows_as_file(model_binary, target_dir)


def load_model(target_dir: str) -> Generator[BlockRow, None, None]:
    model_binary = streaming_tar.build_rows_from_file(target_dir)
    return model_binary


def consume_model(conf: Dict[str, str]):
    # consume the model server to prevent socket server leak.
    # hoverer,  model server may have be consumed by other worker
    # so just try to consume it
    try:
        model_servers = RayContext.parse_servers(conf["modelServers"])
        for item in RayContext.collect_from(model_servers):
            pass
    except Exception as e:
        pass


def common_init_model(
    model_refs: List[ClientObjectRef],
    conf: Dict[str, str],
    model_dir: str,
    is_load_from_local: bool,
):

    udf_name = conf["UDF_CLIENT"] if "UDF_CLIENT" in conf else "UNKNOW MODEL"

    if not is_load_from_local:
        if "standalone" in conf and conf["standalone"] == "true":
            print_flush(
                f"MODEL[{udf_name}] Standalone mode: restore model to {model_dir} directly from model server"
            )
            restore_model(conf, model_dir)
        else:
            print_flush(
                f"MODEL[{udf_name}] Normal mode: restore model from ray object store to {model_dir}"
            )
            if not os.path.exists(model_dir):
                transfer_from_ob(udf_name, model_refs, model_dir)
    else:
        print_flush(
            f"MODEL[{udf_name}]  Local mode: Load model from local path ({model_dir}), consume the model server to prevent socket server leak."
        )
        consume_model(conf)


def parse_params(params: Dict[str, str], prefix: str):
    import json

    new_params = {}
    for k, v in params.items():
        if k.startswith(f"{prefix}."):
            # sft.float.num_train_epochs
            tpe = k.split(".")[1]
            new_k = k.split(".")[2]
            new_v = v
            if tpe == "float":
                new_v = float(v)
            elif tpe == "int":
                new_v = int(v)
            elif tpe == "bool":
                new_v = v == "true"
            elif tpe == "str":
                new_v = v
            elif tpe == "list":
                new_v = json.loads(v)
            elif tpe == "dict":
                new_v = json.loads(v)
            new_params[new_k] = new_v
    return new_params


import inspect


def check_param_exists(func, name):
    return name in inspect.signature(func).parameters


# add a log funcition to log the string to a specified file
def log_to_file(msg: str, file_path: str):
    with open(file_path, "a") as f:
        f.write(msg)
        f.write("\n")


class _PromptWraper:
    def __init__(
        self, func, llm, render, check_result, options, *args, **kwargs
    ) -> None:
        self.func = func
        self.llm = llm
        self.render = render
        self.check_result = check_result
        self.args = args
        self.kwargs = kwargs
        self._options = options

    def options(self, options: Dict[str, Any]):
        self._options = {**self._options, **options}
        return self

    def with_llm(self, llm):
        self.llm = llm
        return self

    def prompt(self):
        func = self.func
        render = self.render
        args = self.args
        kwargs = self.kwargs

        signature = inspect.signature(func)
        arguments = signature.bind(*args, **kwargs)
        arguments.apply_defaults()
        input_dict = {}
        for param in signature.parameters:
            input_dict.update({param: arguments.arguments[param]})

        new_input_dic = func(**input_dict)
        if new_input_dic and not isinstance(new_input_dic, dict):
            raise TypeError(f"Return value of {func.__name__} should be a dict")
        if new_input_dic:
            input_dict = {**input_dict, **new_input_dic}

        if render == "jinja2" or render == "jinja":
            return format_prompt_jinja2(func, **input_dict)

        return format_prompt(func, **input_dict)

    def run(self):
        func = self.func
        llm = self.llm
        render = self.render
        check_result = self.check_result
        args = self.args
        kwargs = self.kwargs

        signature = inspect.signature(func)
        arguments = signature.bind(*args, **kwargs)
        arguments.apply_defaults()
        input_dict = {}
        for param in signature.parameters:
            input_dict.update({param: arguments.arguments[param]})

        is_lambda = inspect.isfunction(llm) and llm.__name__ == "<lambda>"
        if is_lambda:
            if "self" in input_dict:
                instance = input_dict.pop("self")
                return llm(instance).prompt(
                    render=render, check_result=check_result, options=self._options
                )(func)(instance, **input_dict)

        if isinstance(llm, ByzerLLM):
            if "self" in input_dict:
                instance = input_dict.pop("self")
                return llm.prompt(
                    render=render, check_result=check_result, options=self._options
                )(func)(instance, **input_dict)
            else:
                return llm.prompt(
                    render=render, check_result=check_result, options=self._options
                )(func)(**input_dict)

        if isinstance(llm, str):
            _llm = ByzerLLM()
            _llm.setup_default_model_name(llm)
            _llm.setup_template(llm, "auto")

            if "self" in input_dict:
                instance = input_dict.pop("self")
                return _llm.prompt(
                    render=render, check_result=check_result, options=self._options
                )(func)(instance, **input_dict)
            else:
                return _llm.prompt(
                    render=render, check_result=check_result, options=self._options
                )(func)(**input_dict)

        raise ValueError(
            "llm should be a lambda function or ByzerLLM instance or a string of model name"
        )


def prompt_lazy(
    llm=None,
    render: str = "jinja2",
    check_result: bool = False,
    options: Dict[str, Any] = {},
):
    def _impl(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            pw = _PromptWraper(
                func, llm, render, check_result, options, *args, **kwargs
            )
            return pw

        return wrapper

    return _impl


class _PrompRunner:
    def __init__(
        self,
        func,
        instance,
        llm,
        render: str,
        check_result: bool,
        options: Dict[str, Any],
    ) -> None:
        self.func = func
        self.instance = instance
        self.llm = llm
        self.render = render
        self.check_result = check_result
        self._options = options
        self.max_turns = 10
        self.response_markers = None
        self.auto_remove_response_markers_from_output = True
        self.extractor = None
        self.continue_prompt = "接着前面的内容继续"
        self.response_markers_template = """你的输出可能会被切分成多轮对话完成。请确保第一次输出以{{ RESPONSE_START }}开始，输出完毕后，请使用{{ RESPONSE_END }}标记。中间的回复不需要使用这两个标记。"""
        self.model_class = None
        self.return_prefix = None
        self.stop_suffix_list = None

    def with_return_type(self, model_class: Type[Any]):
        self.model_class = model_class
        return self
    
    def with_stop_suffix_list(self, suffix_list: List[str]):
        self.stop_suffix_list = suffix_list
        if "llm_config" in self._options:
            self._options["llm_config"]["gen.stop"] = suffix_list
        else:
            self._options["llm_config"] = {"gen.stop": suffix_list}
        return self

    def with_continue_prompt(self, prompt: str):
        self.continue_prompt = prompt
        return self      

    def with_return_prefix(self, prefix: str):
        self.return_prefix = prefix
        return self

    def __call__(self, *args, **kwargs) -> Any:
        if self.llm:
            return self.run(*args, **kwargs)
        return self.prompt(*args, **kwargs)

    def options(self, options: Dict[str, Any]):
        self._options = {**self._options, **options}
        return self

    def with_response_markers(
        self,
        response_markers: Optional[List[str]] = None,
        response_markers_template: Optional[str] = None,
    ):
        if response_markers is not None and len(response_markers) != 2:
            raise ValueError("response_markers should be a list of two elements")
        if response_markers is None:
            response_markers = ["<RESPONSE>", "</RESPONSE>"]
        self.response_markers = response_markers
        if response_markers_template:
            self.response_markers_template = response_markers_template
        return self

    def with_extractor(self, func):
        self.extractor = func
        return self

    def with_max_turns(self, max_turns: int):
        self.max_turns = max_turns
        return self

    def prompt(self, *args, **kwargs):
        signature = inspect.signature(self.func)
        if self.instance:
            arguments = signature.bind(self.instance, *args, **kwargs)
        else:
            arguments = signature.bind(*args, **kwargs)

        arguments.apply_defaults()
        input_dict = {}
        for param in signature.parameters:
            input_dict.update({param: arguments.arguments[param]})

        new_input_dic = self.func(**input_dict)
        if new_input_dic and not isinstance(new_input_dic, dict):
            raise TypeError(f"Return value of {self.func.__name__} should be a dict")
        if new_input_dic:
            input_dict = {**input_dict, **new_input_dic}

        if "self" in input_dict:
            input_dict.pop("self")

        if self.render == "jinja2" or self.render == "jinja":
            return format_prompt_jinja2(self.func, **input_dict)

        return format_prompt(self.func, **input_dict)

    def with_llm(self, llm):
        self.llm = llm
        return self

    def with_auto_remove_response_markers(self, flag: bool):
        self.auto_remove_response_markers_from_output = flag
        return self

    def _remove_response_markers(self, output: str):
        [start, end] = self.response_markers
        start_index = output.find(start)
        end_index = output.find(end)
        return output[start_index + len(start) : end_index]

    def _multi_turn_wrapper_with_generator(
        self,
        llm,
        v: List[Any],
        signature: inspect.Signature,
        origin_input: Optional[str] = None,
    ):
        conversations = []
        s = ""
        for item in v:
            s += item[0]
            yield item
        if origin_input is None:
            raise ValueError(
                "origin_input should be set when return value is a generator"
            )

        conversations = []
        conversations.append({"role": "user", "content": origin_input})
        conversations.append({"role": "assistant", "content": s})
        turn = 1
        end_marker = self.response_markers[1]
        while turn < self.max_turns and end_marker not in s:
            conversations.append({"role": "user", "content": self.continue_prompt})
            temp_options = {**{"delta_mode": True}, **self._options}
            v1 = llm.stream_chat_oai(conversations=conversations, **temp_options)
            temp = ""
            for item in v1:
                temp += item[0]
                yield item
            conversations.append({"role": "assistant", "content": temp})
            s += temp
            turn += 1
        return

    def _multi_turn_wrapper(
        self,
        llm,
        v: List[Any],
        signature: inspect.Signature,
        origin_input: Optional[str] = None,
    ):

        if not issubclass(signature.return_annotation, str):
            raise ValueError(
                "Return value of function should be a string when response_markers is set"
            )

        conversations = []
        response = v[0]
        s = response.output
        conversations += response.input["history"] or []
        conversations.append({"role": "user", "content": response.input["instruction"]})
        conversations.append({"role": "assistant", "content": response.output})
        turn = 1
        end_marker = self.response_markers[1]

        while turn < self.max_turns and end_marker not in response.output:
            conversations.append({"role": "user", "content": self.continue_prompt})
            v1 = llm.chat_oai(conversations=conversations, **self._options)
            response = v1[0]
            conversations.append({"role": "assistant", "content": response.output})
            s += response.output
            turn += 1

        if self.auto_remove_response_markers_from_output:
            output_content = self._remove_response_markers(output=s)
        response.output = output_content
        if self.extractor:
            if isinstance(self.extractor, str):
                return self.extractor(response)
            else:
                return self.extractor(response.output)
        return response.output

    def is_instance_of_generator(self, v):
        from typing import Generator, get_origin, get_args
        import collections

        if get_origin(v) is collections.abc.Generator:
            args = get_args(v)
            if args == (str, type(None), type(None)):
                return True
        return False

    def to_model(self,result: str):
        from byzerllm.utils.client import code_utils
        from byzerllm.utils.nontext import TagExtractor
                
        if not isinstance(result, str):
            raise ValueError("The decorated function must return a string")
        try:  
            # quick path for json string  
            if result.startswith("```json") and result.endswith("```"):
                json_str = result[len("```json"):-len("```")]
            else:
                json_str = code_utils.extract_code(result)[0][1]
            json_data = json.loads(json_str)            
        except json.JSONDecodeError as e:
            print(f"The returned string is not a valid JSON, e: {str(e)} string: {result}")            
            tag_extractor = TagExtractor(result)
            result = tag_extractor.extract()
            json_data = json.loads(result.content)    
        
        try:
            if isinstance(json_data, list):
                return [self.model_class(**item) for item in json_data]    
            return self.model_class(**json_data)
        except TypeError:
            raise TypeError("Unable to create model instance from the JSON data")

    def run(self, *args, **kwargs):
        func = self.func
        llm = self.llm
        render = self.render
        check_result = self.check_result

        signature = inspect.signature(func)
        if self.instance:
            arguments = signature.bind(self.instance, *args, **kwargs)
        else:
            arguments = signature.bind(*args, **kwargs)

        arguments.apply_defaults()
        input_dict = {}
        for param in signature.parameters:
            input_dict.update({param: arguments.arguments[param]})

        is_lambda = inspect.isfunction(llm) and llm.__name__ == "<lambda>"
        if is_lambda:
            return_origin_response = True if self.response_markers else False
            marker = None

            if self.response_markers:
                marker = format_str_jinja2(
                    self.response_markers_template,
                    RESPONSE_START=self.response_markers[0],
                    RESPONSE_END=self.response_markers[1],
                )

            origin_input = self.prompt(*args, **kwargs)
            if self.response_markers:
                origin_input = f"{origin_input}\n\n{marker}"

            v = llm(self.instance).prompt(
                render=render,
                check_result=check_result,
                options=self._options,
                return_origin_response=return_origin_response,
                marker=marker,
                assistant_prefix=self.return_prefix,
            )(func)(**input_dict)
            prefix = self.return_prefix if self.return_prefix else ""
            if not return_origin_response:                
                if self.extractor:
                    v = self.extractor(f"{prefix}{v}")
                if self.model_class:
                    return self.to_model(f"{prefix}{v}")
                return v

            if self.is_instance_of_generator(signature.return_annotation):
                return self._multi_turn_wrapper_with_generator(
                    llm(self.instance), v, signature, origin_input=origin_input
                )

            v = self._multi_turn_wrapper(
                llm(self.instance), v, signature, origin_input=origin_input
            )
            if self.model_class:
                return self.to_model(f"{prefix}{v}")
            return v

        if isinstance(llm, ByzerLLM):
            return_origin_response = True if self.response_markers else False
            marker = None
            if self.response_markers:
                marker = format_str_jinja2(
                    self.response_markers_template,
                    RESPONSE_START=self.response_markers[0],
                    RESPONSE_END=self.response_markers[1],
                )

            origin_input = self.prompt(*args, **kwargs)
            if self.response_markers:
                origin_input = f"{origin_input}\n\n{marker}"

            v = llm.prompt(
                render=render,
                check_result=check_result,
                options=self._options,
                return_origin_response=return_origin_response,
                marker=marker,
                assistant_prefix=self.return_prefix
            )(func)(**input_dict)
            prefix = self.return_prefix if self.return_prefix else ""
            if not return_origin_response:                
                if self.extractor:
                    v = self.extractor(f"{prefix}{v}")
                if self.model_class:
                    return self.to_model(f"{prefix}{v}")
                return v

            if self.is_instance_of_generator(signature.return_annotation):
                return self._multi_turn_wrapper_with_generator(
                    llm, v, signature, origin_input=origin_input
                )

            v = self._multi_turn_wrapper(llm, v, signature, origin_input=origin_input)            
            return v

        if isinstance(llm, str):
            _llm = ByzerLLM()
            _llm.setup_default_model_name(llm)
            _llm.setup_template(llm, "auto")
            return_origin_response = True if self.response_markers else False
            marker = None

            if self.response_markers:
                marker = format_str_jinja2(
                    self.response_markers_template,
                    RESPONSE_START=self.response_markers[0],
                    RESPONSE_END=self.response_markers[1],
                )

            origin_input = self.prompt(*args, **kwargs)
            if self.response_markers:
                origin_input = f"{origin_input}\n\n{marker}"

            v = _llm.prompt(
                render=render,
                check_result=check_result,
                options=self._options,
                return_origin_response=return_origin_response,
                marker=marker,
                assistant_prefix=self.return_prefix,
            )(func)(**input_dict)
            prefix = self.return_prefix if self.return_prefix else ""
            if not return_origin_response:
                if self.extractor:
                    v = self.extractor(f"{prefix}{v}")
                if self.model_class:
                    return self.to_model(self.model_class)(lambda: f"{prefix}{v}")()
                return v

            if self.is_instance_of_generator(signature.return_annotation):
                return self._multi_turn_wrapper_with_generator(
                    llm, v, signature, origin_input=origin_input
                )

            v = self._multi_turn_wrapper(llm, v, signature, origin_input=origin_input)
            if self.model_class:
                return self.to_model(self.model_class)(lambda: f"{prefix}{v}")()
            return v

        else:
            raise ValueError(
                "llm should be a lambda function or ByzerLLM instance or a string of model name"
            )


class _DescriptorPrompt:
    def __init__(
        self,
        func,
        wrapper,
        llm,
        render: str,
        check_result: bool,
        options: Dict[str, Any],
    ):
        self.func = func
        self.wrapper = wrapper
        self.llm = llm
        self.render = render
        self.check_result = check_result
        self._options = options
        self.prompt_runner = _PrompRunner(
            self.wrapper,
            None,
            self.llm,
            self.render,
            self.check_result,
            options=self._options,
        )

    def __get__(self, instance, owner):
        if instance is None:
            return self
        else:
            return _PrompRunner(
                self.wrapper,
                instance,
                self.llm,
                self.render,
                self.check_result,
                options=self._options,
            )

    def reset(self):
        self.prompt_runner = _PrompRunner(
            self.wrapper,
            None,
            self.llm,
            self.render,
            self.check_result,
            options=self._options,
        )
        return self

    def with_response_markers(
        self,
        response_markers: Optional[List[str]] = None,
        response_markers_template: Optional[str] = None,
    ):
        self.prompt_runner.with_response_markers(
            response_markers=response_markers,
            response_markers_template=response_markers_template,
        )
        return self

    def with_auto_remove_response_markers(self, flag: bool):
        self.prompt_runner.with_auto_remove_response_markers(flag)
        return self

    def with_max_turns(self, max_turns: int):
        self.prompt_runner.with_max_turns(max_turns)
        return self

    def with_continue_prompt(self, prompt: str):
        self.prompt_runner.with_continue_prompt(prompt)
        return self
    
    def with_stop_suffix_list(self, suffix_list: List[str]):
        self.prompt_runner.with_stop_suffix_list(suffix_list)
        return self

    def with_return_type(self, model_class: Type[Any]):
        self.prompt_runner.with_return_type(model_class)
        return self
    
    def with_return_prefix(self, prefix: str):
        self.prompt_runner.with_return_prefix(prefix)
        return self

    def __call__(self, *args, **kwargs):
        return self.prompt_runner(*args, **kwargs)

    def prompt(self, *args, **kwargs):
        return self.prompt_runner.prompt(*args, **kwargs)

    def run(self, *args, **kwargs):
        return self.prompt_runner.run(*args, **kwargs)

    def with_llm(self, llm):
        self.llm = llm
        self.prompt_runner.with_llm(llm)
        return self

    def with_extractor(self, func):
        self.prompt_runner.with_extractor(func)
        return self

    def options(self, options: Dict[str, Any]):
        self._options = {**self._options, **options}
        self.prompt_runner.options(options)
        return self


class prompt:
    def __init__(
        self,
        llm=None,
        render: str = "jinja2",
        check_result: bool = False,
        options: Dict[str, Any] = {},
    ):
        self.llm = llm
        self.render = render
        self.check_result = check_result
        self.options = options

    def __call__(self, func):        
        wrapper = func
        return self._make_wrapper(func, wrapper)

    def _make_wrapper(self, func, wrapper):
        return _DescriptorPrompt(
            func,
            wrapper,
            self.llm,
            self.render,
            self.check_result,
            options=self.options,
        )


from byzerllm.utils.client import ByzerLLM
from byzerllm.utils.retrieval import ByzerRetrieval
from byzerllm.utils.connect_ray import connect_cluster
from byzerllm.apps.agent.registry import reply as agent_reply
from byzerllm.utils.nontext import Image

__all__ = [
    "ByzerLLM",
    "ByzerRetrieval",
    "connect_cluster",
    "prompt",
    "agent_reply",
    "Image",
]



from dataclasses import dataclass
import sys
from typing import Dict, List, Optional, Union
from .agent import Agent
from .conversable_agent import ConversableAgent
import logging
from ray.util.client.common import ClientActorHandle, ClientObjectRef
from ...utils.client import ByzerLLM,code_utils
from byzerllm.utils.retrieval import ByzerRetrieval
import json
from . import get_agent_name, run_agent_func,ChatResponse

try:
    from termcolor import colored
except ImportError:

    def colored(x, *args, **kwargs):
        return x

logger = logging.getLogger(__name__)


@dataclass
class GroupChat:
    """(In preview) A group chat class that contains the following data fields:
    - agents: a list of participating agents.
    - messages: a list of messages in the group chat.
    - max_round: the maximum number of rounds.
    - admin_name: the name of the admin agent if there is one. Default is "Admin".
        KeyBoardInterrupt will make the admin agent take over.
    - func_call_filter: whether to enforce function call filter. Default is True.
        When set to True and when a message is a function call suggestion,
        the next speaker will be chosen from an agent which contains the corresponding function name
        in its `function_map`.
    """

    agents: List[Union[Agent,ClientActorHandle,str]]
    messages: List[Dict]
    max_round: int = 10
    admin_name: str = "Admin"
    func_call_filter: bool = True

    @property
    def agent_names(self) -> List[str]:
        """Return the names of the agents in the group chat."""
        _agent_names = []
        for agent in self.agents:
            _agent_names.append(get_agent_name(agent))
        return _agent_names           


    def reset(self):
        """Reset the group chat."""
        self.messages.clear()

    def agent_by_name(self, name: str) -> Agent:
        """Returns the agent with a given name."""
        return self.agents[self.agent_names.index(name)]    
   

    def next_agent(self, agent: Union[Agent,ClientActorHandle,str], agents: List[Union[Agent,ClientActorHandle,str]]) -> Union[Agent,ClientActorHandle,str]:
        """Return the next agent in the list."""
        agent_name = get_agent_name(agent)
        if agents == self.agents:
            return agents[(self.agent_names.index(agent_name) + 1) % len(agents)]
        else:
            offset = self.agent_names.index(agent_name) + 1
            for i in range(len(self.agents)):
                if self.agents[(offset + i) % len(self.agents)] in agents:
                    return self.agents[(offset + i) % len(self.agents)]

    def select_speaker_msg(self, agents: List[Union[Agent,ClientActorHandle,str]]):
        """Return the message for selecting the next speaker."""
        return f"""You are in a role play game. The following roles are available:
{self._participant_roles()}.

Read the following conversation.
Then select the next role from {[get_agent_name(agent) for agent in agents]} to play. Only return the role."""
     
    

    def select_speaker(self, last_speaker: Union[Agent,ClientActorHandle,str], 
                       selector: Union[ConversableAgent,ClientActorHandle,str],):
        """Select the next speaker."""
        if self.func_call_filter and self.messages and "function_call" in self.messages[-1]:
            # find agents with the right function_map which contains the function name
            agents = [
                agent for agent in self.agents if run_agent_func(agent,"can_execute_function",self.messages[-1]["function_call"]["name"])
            ]
            if len(agents) == 1:
                # only one agent can execute the function
                return agents[0]
            elif not agents:
                # find all the agents with function_map
                agents = [agent for agent in self.agents if run_agent_func(agent,"function_map")]
                if len(agents) == 1:
                    return agents[0]
                elif not agents:
                    raise ValueError(
                        f"No agent can execute the function {self.messages[-1]['name']}. "
                        "Please check the function_map of the agents."
                    )
        else:
            agents = self.agents
            # Warn if GroupChat is underpopulated
            n_agents = len(agents)
            if n_agents < 3:
                logger.warning(
                    f"GroupChat is underpopulated with {n_agents} agents. Direct communication would be more efficient."
                )
        
        run_agent_func(selector,"update_system_message",self.select_speaker_msg(agents))

        select_prompt = self.messages    +        [
                {
                    "role": "user",
                    "content": f"Read the above conversation. Then select the next role from {[get_agent_name(agent) for agent in agents]} to play. Only return the role.",
                }
            ]        
        
        
        final, name = run_agent_func(selector,"generate_llm_reply",None,select_prompt)                                        
        print(colored(f"GroupChat select_speaker: {name}","green"))

        if not final:
            # i = self._random.randint(0, len(self._agent_names) - 1)  # randomly pick an id
            return self.next_agent(last_speaker, agents)
        try:
            return self.agent_by_name(name.strip())
        except ValueError:
            logger.warning(
                f"GroupChat select_speaker failed to resolve the next speaker's name. Speaker selection will default to the next speaker in the list. This is because the speaker selection OAI call returned:\n{name}"
            )
            return self.next_agent(last_speaker, agents)

    def _participant_roles(self):
        roles = []
        for agent in self.agents:
            if run_agent_func(agent,"get_system_message").strip() == "":
                logger.warning(
                    f"The agent '{get_agent_name(agent)}' has an empty system_message, and may not work well with GroupChat."
                )
            roles.append(f"{get_agent_name(agent)}: {run_agent_func(agent,'get_system_message')}")
        return "\n".join(roles)


class GroupChatManager(ConversableAgent):
    """(In preview) A chat manager agent that can manage a group chat of multiple agents."""

    def __init__(
        self,
        groupchat: GroupChat,
        llm:ByzerLLM,
        retrieval:ByzerRetrieval,
        name: Optional[str] = "chat_manager",        
        # unlimited consecutive auto reply by default
        max_consecutive_auto_reply: Optional[int] = sys.maxsize,
        human_input_mode: Optional[str] = "NEVER",
        system_message: Optional[str] = "Group chat manager.",
        **kwargs,
    ):
        super().__init__(
            name=name,
            llm=llm,
            retrieval=retrieval,
            max_consecutive_auto_reply=max_consecutive_auto_reply,
            human_input_mode=human_input_mode,
            system_message=system_message,
            **kwargs,
        )
        # Order of register_reply is important.
        # Allow sync chat if initiated using initiate_chat
        self.register_reply([Agent, ClientActorHandle,str], 
                            GroupChatManager.run_chat, 
                            config=groupchat, 
                            reset_config=GroupChat.reset)  
        self.groupchat = groupchat  

    def get_groupchat(self) -> GroupChat:
        """Return the group chat."""
        return self.groupchat   

    def reset_agents(self):
        """Reset the agents."""
        for agent in self.groupchat.agents:
            run_agent_func(agent,"reset")       

    def run_chat(
        self,
        raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Union[Agent, ClientActorHandle,str]] = None,
        config: Optional[GroupChat] = None,
    ) -> Union[str, Dict, None]:
        """Run a group chat."""
        if messages is None:
            messages = self._messages[get_agent_name(sender)]
        message = messages[-1]
        speaker = sender
        groupchat = config
        for i in range(groupchat.max_round):
            print(colored(f"GroupChatManager run_chat: {i}","green"),flush=True)
            # set the name to speaker's name if the role is not function
            if message["role"] != "function":
                message["name"] = get_agent_name(speaker)
            groupchat.messages.append(message)
            # broadcast the message to all agents except the speaker
            for agent in groupchat.agents:
                if get_agent_name(agent) != get_agent_name(speaker):
                    self.send(message, agent, request_reply=False, silent=True)
            if i == groupchat.max_round - 1:
                # the last round
                break
            try:
                # select the next speaker
                speaker = groupchat.select_speaker(speaker, self)
                # let the speaker speak
                reply = run_agent_func(speaker,"generate_reply",sender=self)
            except KeyboardInterrupt:
                # let the admin agent speak if interrupted
                if groupchat.admin_name in groupchat.agent_names:
                    # admin agent is one of the participants
                    speaker = groupchat.agent_by_name(groupchat.admin_name)
                    reply = run_agent_func(speaker,"generate_reply",sender=self)
                else:
                    # admin agent is not found in the participants
                    raise
            if reply is None:
                break
            # The speaker sends the message without requesting a reply
            run_agent_func(speaker,"send",message=reply,recipient=get_agent_name(self),request_reply=False);
            # get the speaker's last message and in next round, broadcast it to all other agents            
            message = self.last_message(speaker)            
        return True, None
    



import json
import os

import deepspeed
import deepspeed.comm as dist
import numpy as np
import sentencepiece as spm
import torch
from torch.utils.data import Dataset
from pyjava import RayContext

from transformers import Trainer, TrainingArguments,default_data_collator

from .base_model.configuration_baichuan import BaiChuanConfig
from .base_model.modeling_baichuan import BaiChuanForCausalLM
from . import TrainParameters
from ray.train.huggingface import TransformersTrainer
from ray.air.config import ScalingConfig
from ray.air import session

deepspeed_confg = json.loads('''
    {
  "gradient_accumulation_steps": 1,
  "train_micro_batch_size_per_gpu": 1,
  "prescale_gradients": false,
  "zero_allow_untested_optimizer": true,
  "optimizer": {
    "type": "AdamW",
    "params": {
      "lr": 1e-8,
      "eps": 1.0e-8,
      "betas": [
        0.9,
        0.95
      ],
      "weight_decay": 0.1
    }
  },
  "tensorboard": {
    "enabled": true,
    "output_path": "logs/",
    "job_name": "baichuan-7b-pt"
  },
  "zero_optimization": {
    "stage": 3,
    "contiguous_gradients": false,
    "allgather_bucket_size": 3e8,
    "reduce_bucket_size": 3e8,
    "overlap_comm": true,
    "reduce_scatter": true
  },
  "steps_per_print": 16,
  "gradient_clipping": 1.0,
  "wall_clock_breakdown": true,
  "bf16": {
    "enabled": true
  }
}''')
                             
class DataEngine():
    def __init__(self, data_dir, tokenizer_path, micro_batch_size, max_length):
        self.MIN_TEXT_LEN = 20
        self.EOS_TOKEN_ID = 2
        self.data_dir = data_dir
        self.sp = spm.SentencePieceProcessor()
        self.sp.Load(tokenizer_path)
        self.micro_batch_size = micro_batch_size
        self.max_length = max_length
        self.data = []
        self.global_input_paths = [self.data_dir + "/" + x
                                   for x in os.listdir(self.data_dir)]
        self.local_input_paths = [x for i, x in enumerate(self.global_input_paths)]

    def load_data(self):
        for file_path in self.local_input_paths:
            data = []
            with open(file_path, encoding="utf-8", errors="ignore") as f:
                for line_id, line in enumerate(f):
                    line_json = json.loads(line.strip())
                    text = line_json("instruction") + line_json("input") + line_json("output")
                    cc = self.sp.EncodeAsIds(text) + [self.EOS_TOKEN_ID]
                    if len(cc) < self.MIN_TEXT_LEN:
                        cc = []
                    data.extend(cc)
                    if len(data) >= self.micro_batch_size * (self.max_length + 1):
                        index = self.micro_batch_size * (self.max_length + 1)
                        self.data.append(data[:index])
                        data = []
        return

    def get_data(self):
        data = self.data.pop(0)
        seq = np.asarray(data).reshape(self.micro_batch_size, self.max_length + 1)
        data = torch.LongTensor(seq)
        # data = data.cuda(non_blocking=True)
        return data                             
                             
class FulltuneDataset(Dataset):

    def __init__(self,data_engine:DataEngine) -> None:
        self.data_engine = data_engine


    def __len__(self):
        return len(self.data_engine.data)

    def __getitem__(self, index):                
        input_ids = self.data_engine.get_data()
        inputs = {
            'input_ids': input_ids            
        }
        return inputs                       


def trainer_init_per_worker(train_dataset, eval_dataset=None, **config):
    # Use the actual number of CPUs assigned by Ray
    os.environ["OMP_NUM_THREADS"] = str(
        session.get_trial_resources().bundles[-1].get("CPU", 1)
    )
    # Enable tf32 for better performance
    # torch.backends.cuda.matmul.allow_tf32 = True

    batch_size = config.get("batch_size", 4)
    epochs = config.get("epochs", 2)
    warmup_steps = config.get("warmup_steps", 0)
    learning_rate = config.get("learning_rate", 0.00002)
    weight_decay = config.get("weight_decay", 0.01)

    print("Preparing training arguments")
    training_args = TrainingArguments(
        "output",
        per_device_train_batch_size=batch_size,
        logging_steps=1,
        save_strategy="no",
        per_device_eval_batch_size=batch_size,
        learning_rate=learning_rate,
        weight_decay=weight_decay,
        warmup_steps=warmup_steps,
        label_names=["input_ids", "attention_mask"],
        num_train_epochs=epochs,
        push_to_hub=False,
        disable_tqdm=True,  # declutter the output a little
        fp16=True,
        gradient_checkpointing=True,
        deepspeed=deepspeed_confg,
    )
    
    data_refs = config["data_refs"]
    train_args = config["train_args"] 
    index = 0 # dist.get_rank()
        
    if not os.path.exists(os.path.join(train_args.data_dir,str(index))):
        os.makedirs(os.path.join(train_args.data_dir,str(index)))
    
    data_file_path = os.path.join(train_args.data_dir,str(index),"data.json")
    with open(data_file_path,"w",encoding="utf-8") as f:        
        for item in RayContext.collect_from([data_refs[index]]):
            f.write(json.dumps(item,ensure_ascii=False)+"\n")
                                
    micro_batch_size =  deepspeed_confg["train_micro_batch_size_per_gpu"]
    data_engine = DataEngine(train_args.data_dir, train_args.tokenizer_path, micro_batch_size, train_args.max_length)
    data_engine.load_data()   
    fulltune_dataset = FulltuneDataset(data_engine) 

    os.remove(data_file_path)

    model = BaiChuanForCausalLM(BaiChuanConfig())

    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=fulltune_dataset        
    )
    return trainer



def distribute_train(args:TrainParameters,data_refs):

    assert  args.num_workers == len(data_refs), f"num_workers({args.num_workers}) must equal to data_refs({len(data_refs)})"    

    distribute_trainer = TransformersTrainer(
                                trainer_init_per_worker=trainer_init_per_worker,
                                trainer_init_config={
                                    "batch_size": 16,  # per device
                                    "epochs": 1,
                                    "data_refs":data_refs,
                                    "train_args":args
                                },
                                scaling_config=ScalingConfig(
                                    num_workers=args.num_workers,
                                    use_gpu=args.use_gpu,
                                    resources_per_worker={"GPU": args.gpus_per_worker, "CPU": args.cpus_per_worker},
                                )                                                                                               
    ) 
    r = distribute_trainer.fit()
    return r.path()




import whisper
from typing import Dict, List, Tuple
from byzerllm.utils.types import StopSequencesCriteria

from typing import Dict, Any, List, Union
from pyjava.storage import streaming_tar as STar
import json
import base64
import tempfile
import time


def get_meta(self):
    return [
        {
            "model_deploy_type": "proprietary",
            "backend": "transformers",
            "message_format": True
        }
    ]


def process_input(ins: Union[str, List[Dict[str, Any]], Dict[str, Any]]):
    # print(ins) 
    if isinstance(ins, list) or isinstance(ins, dict):
        return ins

    content = []
    try:
        ins_json = json.loads(ins)
    except:
        return ins

    ## speech
    if isinstance(ins_json, dict):
        return ins_json

    content = []
    #     [
    #     {"type": "text", "text": "What’s in this image?"},
    #     {
    #       "type": "image_url",
    #       "image_url": {
    #         "url": "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg",
    #         "detail": "high"
    #       },
    #     },
    #   ],
    for item in ins_json:
        # for format like this: {"image": "xxxxxx", "text": "What’s in this image?","detail":"high"}
        # or {"image": "xxxxxx"}, {"text": "What’s in this image?"}
        if ("image" in item or "image_url" in item) and "type" not in item:
            image_data = item.get("image", item.get("image_url", ""))
            ## "data:image/jpeg;base64,"
            if not image_data.startswith("data:"):
                image_data = "data:image/jpeg;base64," + image_data

            ## get the other fields except image/text/image_url
            other_fields = {
                k: v for k, v in item.items() if k not in ["image", "text", "image_url"]
            }
            content.append(
                {
                    "image_url": {"url": image_data, **other_fields},
                    "type": "image_url",
                }
            )

        if "text" in item and "type" not in item:
            text_data = item["text"]
            content.append({"text": text_data, "type": "text"})

        ## for format like this: {"type": "text", "text": "What’s in this image?"},
        ## {"type": "image_url", "image_url": {"url":"","detail":"high"}}
        ## this is the standard format, just return it
        if "type" in item and item["type"] == "text":
            content.append(item)

        if "type" in item and item["type"] == "image_url":
            content.append(item)

    if not content:
        return ins

    return content


def stream_chat(
    self,
    tokenizer,
    ins: str,
    his: List[Dict[str, str]] = [],
    max_length: int = 4090,
    top_p: float = 0.95,
    temperature: float = 0.1,
    **kwargs,
):
    messages = [
        {"role": message["role"], "content": process_input(message["content"])}
        for message in his
    ] + [{"role": "user", "content": process_input(ins)}]

    audio_input = messages[-1]["content"]

    audio_data = audio_input.get("audio", "")
    audio = audio_input["audio"]
    response_format = audio_input.get("response_format", "json")
    timestamp_granularities = audio_input.get(
        "timestamp_granularities", ["word", "segment"]
    )
    data_prefix = "data:audio/"
    base64_prefix = ";base64,"
    if not audio.startswith(data_prefix):
        raise ValueError("Invalid audio data format")

    format_end = audio.index(base64_prefix)
    audio_format = audio[len(data_prefix) : format_end]
    base64_data = audio[format_end + len(base64_prefix) :]

    # Decode the base64 audio data
    audio_data = base64.b64decode(base64_data)

    # Create a temporary file with the correct extension
    with tempfile.NamedTemporaryFile(
        delete=False, suffix=f".{audio_format}"
    ) as temp_audio_file:
        temp_audio_file.write(audio_data)
        temp_audio_file_path = temp_audio_file.name

    try:
        start_time = time.monotonic()
        word_timestamps = "word" in timestamp_granularities
        
        result = self.transcribe(temp_audio_file_path, word_timestamps=word_timestamps)        
            
        time_cost = time.monotonic() - start_time
        return [
            (
                json.dumps(result, ensure_ascii=False),
                {
                    "metadata": {
                        "request_id": "",
                        "input_tokens_count": 0,
                        "generated_tokens_count": 0,
                        "time_cost": time_cost,
                        "first_token_time": 0,
                        "speed": 0,
                    }
                },
            )
        ]
    finally:
        # Clean up the temporary file
        import os

        os.unlink(temp_audio_file_path)


def init_model(
    model_dir, infer_params: Dict[str, str] = {}, sys_conf: Dict[str, str] = {}
):
    model = whisper.load_model("large", download_root=model_dir)
    import types

    model.stream_chat = types.MethodType(stream_chat, model)
    model.get_meta = types.MethodType(get_meta, model)
    return (model, None)



import time
from typing import List, Tuple, Dict, Any, Union
import httpx
import requests
import base64
import io
import json
import requests
import ray
from byzerllm.utils.types import (
    BlockVLLMStreamServer,
    StreamOutputs,
    SingleOutput,
    SingleOutputMeta,
    BlockBinaryStreamServer,
)
from byzerllm.utils.langutil import asyncfy_with_semaphore
import threading
import asyncio
import traceback
import uuid
import tempfile
from loguru import logger


class CustomSaasAPI:

    def __init__(self, infer_params: Dict[str, str]) -> None:

        self.api_key = infer_params["saas.api_key"]
        self.model = infer_params.get("saas.model", "black-forest-labs/FLUX.1-dev")

        self.meta = {
            "model_deploy_type": "saas",
            "backend": "saas",
            "support_stream": False,
            "model_name": self.model,
        }

        try:
            ray.get_actor("BLOCK_VLLM_STREAM_SERVER")
        except ValueError:
            try:
                ray.remote(BlockVLLMStreamServer).options(
                    name="BLOCK_VLLM_STREAM_SERVER",
                    lifetime="detached",
                    max_concurrency=1000,
                ).remote()
            except Exception as e:
                pass
        try:
            ray.get_actor("BlockBinaryStreamServer")
        except ValueError:
            try:
                ray.remote(BlockBinaryStreamServer).options(
                    name="BlockBinaryStreamServer",
                    lifetime="detached",
                    max_concurrency=1000,
                ).remote()
            except Exception as e:
                pass

    def get_meta(self):
        return [self.meta]

    def process_input(self, ins: Union[str, List[Dict[str, Any]], Dict[str, Any]]):
        if isinstance(ins, list) or isinstance(ins, dict):
            return ins
        try:
            return json.loads(ins)
        except:
            return ins

    async def async_text_to_image(
        self, stream: bool, input: str, size: str, quality: str, n: int, **kwargs
    ):
        if stream:
            raise Exception("Stream not supported for text to image")

        start_time = time.monotonic()
        url = "https://api.siliconflow.cn/v1/image/generations"

        payload = {
            "model": self.model,
            "prompt": input,
            "image_size": size,
            "num_inference_steps": kwargs.get("num_inference_steps", 20),
        }
        headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "Authorization": f"Bearer {self.api_key}",
        }

        response = await asyncfy_with_semaphore(
            lambda: requests.post(url, json=payload, headers=headers)
        )()
        response_data = response.json()

        time_cost = time.monotonic() - start_time
        
        if "images" not in response_data:
            print(response_data)
            raise Exception(response_data)
        
        image_url = response_data["images"][0]["url"]
        image_response = await asyncfy_with_semaphore(lambda: requests.get(image_url))()
        image_binary = image_response.content
        # base64_image = f"data:image/png;base64,{base64.b64encode(image_binary).decode('utf-8')}"
        base64_image = base64.b64encode(image_binary).decode("utf-8")

        return [
            (
                base64_image,
                {
                    "metadata": {
                        "request_id": "",
                        "input_tokens_count": 0,
                        "generated_tokens_count": 0,
                        "time_cost": time_cost,
                        "first_token_time": 0,
                        "speed": 0,
                    }
                },
            )
        ]

    async def async_stream_chat(
        self,
        tokenizer,
        ins: str,
        his: List[Dict[str, Any]] = [],
        max_length: int = 4096,
        top_p: float = 0.7,
        temperature: float = 0.9,
        **kwargs,
    ):
        logger.info(f"[{self.model}] request accepted: {ins[-50:]}....")

        messages = [
            {"role": message["role"], "content": self.process_input(message["content"])}
            for message in his
        ] + [{"role": "user", "content": self.process_input(ins)}]

        last_message = messages[-1]["content"]

        if isinstance(last_message, dict) and "input" in last_message:
            input = last_message["input"]
            size = last_message.get("size", "1024x1024")
            quality = last_message.get("quality", "standard")
            n = last_message.get("n", 1)
            # Get other parameters excluding input, size, quality, and n
            other_params = {
                k: v
                for k, v in last_message.items()
                if k not in ["input", "size", "quality", "n"]
            }

            return await self.async_text_to_image(
                stream=False,
                input=input,
                size=size,
                quality=quality,
                n=n,
                **other_params,
            )
        raise Exception("Invalid input")



from ray.util.placement_group import (    
    remove_placement_group,
    PlacementGroup
)
from ray._private.utils import hex_to_binary
from ray._raylet import PlacementGroupID
from ray.util.state import (
        StateApiClient,
        get_log,
        list_logs,
        summarize_actors,
        summarize_objects,
        summarize_tasks,
    )
from ray.util.state.common import (
    DEFAULT_LIMIT,
    DEFAULT_LOG_LIMIT,
    DEFAULT_RPC_TIMEOUT,
    GetApiOptions,
    ListApiOptions,
    PredicateType,
    StateResource,
    StateSchema,
    SupportedFilterType,
    resource_to_schema,
)
from ray.util.state.exception import RayStateApiException
from ray.util.annotations import PublicAPI
import ray

def cancel_placement_group(group_id:str):
    remove_placement_group(PlacementGroup(
                PlacementGroupID(hex_to_binary(group_id))
            ))

def get_actor_info(actor):            
    resource = StateResource("actors".replace("-", "_"))
    # Create the State API server and put it into context
    client = StateApiClient(address="auto")
    options = GetApiOptions(timeout=30)
    # If errors occur, exceptions will be thrown. Empty data indicate successful query.
    try:
        state = client.get(
            resource,
            options=options,
            id=actor._ray_actor_id.hex(),
            _explain=True,
        )
        return state
    except RayStateApiException as e:
        raise e

def is_ray_in_client_mode():
    if ray.is_initialized() and ray.util.client.ray.is_connected():
        return True
    return False    
    

    
  
    



import pydantic
from typing import Any, Dict
from abc import ABC, abstractmethod


class Message(pydantic.BaseModel):
    id: str
    m: Dict[str, Any]
    sender: str
    receiver: str
    timestamp: float


class MessageStore(ABC):    
    @abstractmethod
    def put(self, message: Message):
        pass
    
    @abstractmethod
    def get(self, id: str):
        pass

    @abstractmethod
    def clear(self, id: str):
        pass



import asyncio
import json
import threading
import time
import traceback
from typing import List, Dict, Any, Union
import ray
from anthropic import Anthropic

from byzerllm.utils.types import (
    BlockVLLMStreamServer,
    StreamOutputs,
    SingleOutput,
    SingleOutputMeta,
)
from byzerllm.utils.langutil import asyncfy_with_semaphore


class CustomSaasAPI:
    def __init__(self, infer_params: Dict[str, str]) -> None:
        self.api_key: str = infer_params["saas.api_key"]
        self.model = infer_params.get("saas.model", "claude-3-haiku-20240307")
        self.meta = {
            "model_deploy_type": "saas",
            "backend": "saas",
            "support_stream": True,
            "support_assistant_prefix": True,
        }
        other_params = {}

        if "saas.base_url" in infer_params:
            other_params["base_url"] = infer_params["saas.base_url"]

        self.client = Anthropic(api_key=self.api_key, **other_params)

        try:
            ray.get_actor("BLOCK_VLLM_STREAM_SERVER")
        except ValueError:
            try:
                ray.remote(BlockVLLMStreamServer).options(
                    name="BLOCK_VLLM_STREAM_SERVER",
                    lifetime="detached",
                    max_concurrency=1000,
                ).remote()
            except:
                pass

    # saas/proprietary
    def get_meta(self):
        return [self.meta]

    async def async_get_meta(self):
        return await asyncfy_with_semaphore(self.get_meta)()

    def process_input(self, ins: Union[str, List[Dict[str, Any]], Dict[str, Any]]):
        if isinstance(ins, list) or isinstance(ins, dict):
            return ins

        content = []
        try:
            ins_json = json.loads(ins)
        except:
            return ins

        if isinstance(ins_json, dict):
            return ins_json

        content = []
        for item in ins_json:
            if "image" in item or "image_url" in item and "type" not in item:
                image_data = item.get("image", item.get("image_url", ""))
                if not image_data.startswith("data:"):
                    image_data = "data:image/jpeg;base64," + image_data

                data_prefix = "data:image/"
                base64_prefix = ";base64,"
                if not image_data.startswith(data_prefix):
                    raise ValueError("Invalid image data format")

                format_end = image_data.index(base64_prefix)
                image_format = image_data[len(data_prefix) : format_end]
                base64_data = image_data[format_end + len(base64_prefix) :]

                content.append(
                    {
                        "type": "image",
                        "source": {
                            "type": "base64",
                            "media_type": f"image/{image_format}",
                            "data": base64_data,
                        },
                    }
                )

            if "text" in item and "type" not in item:
                text_data = item["text"]
                content.append({"type": "text", "text": text_data})

            if "type" in item and item["type"] in ["text"]:
                content.append(item)
            # 兼容openai {"type": "image_url", "image_url": {"url":"","detail":"high"}}
            if "type" in item and item["type"] == "image_url":
                image_data = item["image_url"]["url"]
                data_prefix = "data:image/"
                base64_prefix = ";base64,"
                if not image_data.startswith(data_prefix):
                    raise ValueError("Invalid image data format")

                format_end = image_data.index(base64_prefix)
                image_format = image_data[len(data_prefix) : format_end]
                base64_data = image_data[format_end + len(base64_prefix) :]
                content.append(
                    {
                        "type": "image",
                        "source": {
                            "type": "base64",
                            "media_type": f"image/{image_format}",
                            "data": base64_data,
                        },
                    }
                )

        if not content:
            return ins

        return content

    async def async_stream_chat(
        self,
        tokenizer,
        ins: str,
        his: List[dict] = [],
        max_length: int = 1024,
        top_p: float = 0.9,
        temperature: float = 0.1,
        **kwargs,
    ):
        messages = []
        system_message = ""
        for message in his:
            if message["role"] == "system":
                system_message = message["content"]
            else:
                messages.append(
                    {
                        "role": message["role"],
                        "content": self.process_input(message["content"]),
                    }
                )

        messages.append({"role": "user", "content": self.process_input(ins)})
        
        # 支持assistant prefix
        if len(messages) > 1 and messages[-1]["role"] == "user" and messages[-2]["role"] == "user":
            messages[-1]["role"] = "assistant"

        start_time = time.monotonic()

        other_params = {}

        if system_message:
            other_params["system"] = system_message

        if "stream" in kwargs:
            other_params["stream"] = kwargs["stream"]

        stream = kwargs.get("stream", False)

        try:
            res_data = await asyncfy_with_semaphore(
                lambda: self.client.messages.create(
                    model=self.model,
                    max_tokens=max_length,
                    temperature=temperature,
                    top_p=top_p,
                    messages=messages,
                    **other_params,
                )
            )()
        except Exception as e:
            traceback.print_exc()
            print(messages)
            raise e

        if stream:
            server = ray.get_actor("BLOCK_VLLM_STREAM_SERVER")
            request_id = [None]

            def writer():
                input_tokens = 0
                r = ""
                for response in res_data:

                    if response.type == "message_start":
                        request_id[0] = response.message.id
                        input_tokens = response.message.usage.input_tokens

                    if response.type == "content_block_delta":
                        v = response.delta.text
                        r += v
                        server.add_item.remote(
                            request_id[0],
                            StreamOutputs(
                                outputs=[
                                    SingleOutput(
                                        text=r,
                                        metadata=SingleOutputMeta(
                                            input_tokens_count=0,
                                            generated_tokens_count=0,
                                        ),
                                    )
                                ]
                            ),
                        )
                    if response.type == "message_delta":
                        server.add_item.remote(
                            request_id[0],
                            StreamOutputs(
                                outputs=[
                                    SingleOutput(
                                        text=r,
                                        metadata=SingleOutputMeta(
                                            input_tokens_count=input_tokens,
                                            generated_tokens_count=response.usage.output_tokens,
                                        ),
                                    )
                                ]
                            ),
                        )

                server.mark_done.remote(request_id[0])

            threading.Thread(target=writer, daemon=True).start()

            time_count = 10 * 100
            while request_id[0] is None and time_count > 0:
                await asyncio.sleep(0.01)
                time_count -= 1

            if request_id[0] is None:
                raise Exception("Failed to get request id")

            def write_running():
                return ray.get(server.add_item.remote(request_id[0], "RUNNING"))

            await asyncio.to_thread(write_running)
            return [
                (
                    "",
                    {
                        "metadata": {
                            "request_id": request_id[0],
                            "stream_server": "BLOCK_VLLM_STREAM_SERVER",
                        }
                    },
                )
            ]

        time_cost = time.monotonic() - start_time

        generated_text = res_data.content[0].text
        generated_tokens_count = res_data.usage.output_tokens
        input_tokens_count = res_data.usage.input_tokens

        return [
            (
                generated_text,
                {
                    "metadata": {
                        "request_id": res_data.id,
                        "input_tokens_count": input_tokens_count,
                        "generated_tokens_count": generated_tokens_count,
                        "time_cost": time_cost,
                        "first_token_time": 0,
                        "speed": float(generated_tokens_count) / time_cost,
                        "stop_reason": res_data.stop_reason,
                    }
                },
            )
        ]



# Adapted from
# vLLM project

import asyncio
import time
from typing import AsyncGenerator, Optional

from fastapi import Request

from byzerllm.utils.types import SingleOutputMeta
from byzerllm.utils import random_uuid
from byzerllm.utils.client import ByzerLLM, LLMResponse
from byzerllm.utils.client.entrypoints.openai.protocol import (
    CompletionRequest,
    CompletionResponse,
    CompletionResponseChoice,
    CompletionResponseStreamChoice,
    CompletionStreamResponse,
    LogProbs,
    UsageInfo,
)
from byzerllm.utils.client.entrypoints.openai.serving_engine import OpenAIServing


class OpenAIServingCompletion(OpenAIServing):

    def __init__(
            self,
            llm_client: ByzerLLM,
            server_model_name: Optional[str] = None,
            prompt_template: Optional[str] = None
    ):
        super().__init__(
            llm_client=llm_client,
            server_model_name=server_model_name,
            prompt_template=prompt_template
        )

    async def create_completion(
            self,
            body: CompletionRequest,
            request: Request
    ):
        """Completion API similar to OpenAI's API.

        See https://platform.openai.com/docs/api-reference/completions/create
        for the API specification. This API mimics the OpenAI Completion API.

        NOTE: Currently we do not support the following feature:
            - suffix (the language models we currently support do not support
            suffix)
        """

        # OpenAI API supports echoing the prompt when max_tokens is 0.
        echo_without_generation = body.echo and body.max_tokens == 0

        if body.prompt_template:
            self.llm_client.setup_template(body.model, self._detect_prompt_template(body.prompt_template))

        model_name = self.server_model_name or body.model
        request_id = f"cmpl-{random_uuid()}"
        created_time = int(time.monotonic())

        # Similar to the OpenAI API, when n != best_of, we do not stream the
        # results. In addition, we do not stream the results when use beam search.
        if body.stream:
            return self.completion_stream_generator(body, request_id, created_time)

        # Non-streaming response
        async def wrapper_chat_generator():
            r = self.llm_client.chat_oai(
                model=model_name,
                conversations=[
                    {
                        "role": "user",
                        "content": body.prompt
                    }
                ],
                llm_config={
                    "gen.request_id": request_id,
                    **body.to_llm_config()
                }
            )
            for _ in r:
                yield _

        result_generator = await asyncio.to_thread(wrapper_chat_generator)
        final_res = None
        async for res in result_generator:
            if await request.is_disconnected():
                # Abort the request if the client disconnects.
                await self.llm_client.abort(request_id, model=model_name)
                return self.create_error_response("Client disconnected")
            final_res = res
        assert final_res is not None
        choices = []

        for r in [final_res]:
            r: LLMResponse
            choice_data = CompletionResponseChoice(
                index=0,
                text=r.output,
                logprobs=None,
                finish_reason=None,
            )
            choices.append(choice_data)

        num_prompt_tokens = r.metadata.get("input_tokens_count", 0)
        num_generated_tokens = r.metadata.get("generated_tokens_count", 0)
        usage = UsageInfo(
            prompt_tokens=num_prompt_tokens,
            completion_tokens=num_generated_tokens,
            total_tokens=num_prompt_tokens + num_generated_tokens,
        )
        response = CompletionResponse(
            id=request_id,
            created=created_time,
            model=model_name,
            choices=choices,
            usage=usage,
        )

        if body.stream:
            # When user requests streaming but we don't stream, we still need to
            # return a streaming response with a single event.
            response_json = response.json(ensure_ascii=False)

            async def fake_stream_generator() -> AsyncGenerator[str, None]:
                yield f"data: {response_json}\n\n"
                yield "data: [DONE]\n\n"

            return fake_stream_generator()

        return response

    def create_stream_response_json(
            self,
            body: CompletionRequest,
            index: int,
            text: str,
            request_id: str,
            created_time: int,
            logprobs: Optional[LogProbs] = None,
            finish_reason: Optional[str] = None,
            usage: Optional[UsageInfo] = None,
    ) -> str:
        choice_data = CompletionResponseStreamChoice(
            index=index,
            text=text,
            logprobs=logprobs,
            finish_reason=finish_reason,
        )
        response = CompletionStreamResponse(
            id=request_id,
            created=created_time,
            model=body.model,
            choices=[choice_data],
        )
        if usage is not None:
            response.usage = usage
        response_json = response.model_dump_json(exclude_unset=True)

        return response_json

    async def completion_stream_generator(
            self,
            body: CompletionRequest,
            request_id: str,
            created_time: int
    ) -> AsyncGenerator[str, None]:
        model_name = self.server_model_name or body.model
        previous_texts = [""] * body.n
        result_generator = self.llm_client.async_stream_chat_oai(
            model=model_name,
            conversations=[{
                "role": "user",
                "content": body.prompt
            }],
            delta_mode=True,
            llm_config={
                "gen.request_id": request_id,
                **body.to_llm_config()
            }
        )

        async for res in result_generator:
            (s, meta) = res
            meta: SingleOutputMeta
            for _ in [(s, meta)]:
                i = 0
                delta_text = s[len(previous_texts[i]):]
                top_logprobs = None
                logprobs = None

                previous_texts[i] = s
                finish_reason = None
                response_json = self.create_stream_response_json(
                    body=body,
                    index=i,
                    text=delta_text,
                    logprobs=logprobs,
                    request_id=request_id,
                    created_time=created_time,
                    finish_reason=finish_reason,
                )
                yield f"data: {response_json}\n\n"
                completion_tokens = meta.generated_tokens_count
                prompt_tokens = meta.input_tokens_count
                final_usage = UsageInfo(
                    prompt_tokens=prompt_tokens,
                    completion_tokens=completion_tokens,
                    total_tokens=prompt_tokens + completion_tokens,
                )
                response_json = self.create_stream_response_json(
                    body=body,
                    index=i,
                    text="",
                    request_id=request_id,
                    created_time=created_time,
                    logprobs=logprobs,
                    finish_reason=None,
                    usage=final_usage,
                )
                yield f"data: {response_json}\n\n"
        yield "data: [DONE]\n\n"



from zhipuai import ZhipuAI
import time
from typing import List, Tuple, Dict,Any
import ray
from byzerllm.utils.types import BlockVLLMStreamServer,StreamOutputs,SingleOutput,SingleOutputMeta
from byzerllm.utils.langutil import asyncfy_with_semaphore
import threading
import asyncio


class CustomSaasAPI:
    def __init__(self, infer_params: Dict[str, str]) -> None:
        self.api_key = infer_params["saas.api_key"]
        # chatglm_lite, chatglm_std, chatglm_pro
        self.model = infer_params.get("saas.model", "glm-4")        
        self.client = ZhipuAI(api_key=self.api_key) 

        self.meta = {
            "model_deploy_type": "saas",
            "backend":"saas",
            "support_stream": True
        }

        if "embedding" not in  self.model.lower():            
            self.meta["embedding_mode"] = False 
        else:            
            self.meta["embedding_mode"] = True       
        try:
            ray.get_actor("BLOCK_VLLM_STREAM_SERVER")
        except ValueError:            
            ray.remote(BlockVLLMStreamServer).options(name="BLOCK_VLLM_STREAM_SERVER",lifetime="detached",max_concurrency=1000).remote()            

    # saas/proprietary
    def get_meta(self):
        return [self.meta]

    async def async_get_meta(self):
        return await asyncfy_with_semaphore(self.get_meta)() 

    def embed_query(self, ins: str, **kwargs):                     
        start_time = time.monotonic()
        response = self.client.embeddings.create(
                model=self.model,
                input=ins,
            )
        time_cost = time.monotonic() - start_time
        return response.data[0].embedding
    
    async def async_embed_query(self, ins: str, **kwargs):
        return await asyncfy_with_semaphore(self.embed_query)(ins, **kwargs)

    async def async_stream_chat(self, tokenizer, ins: str, his: List[Dict[str, Any]] = [],
                    max_length: int = 4096,
                    top_p: float = 0.7,
                    temperature: float = 0.9, **kwargs):
        
        messages = his + [{"role": "user", "content": ins}]
        
        stream = kwargs.get("stream",False)    

        other_params = {}
        
        if "stream" in kwargs:        
            other_params["stream"] = kwargs["stream"]

        for k, v in kwargs.items():
            if k in ["max_tokens", "stop"]:
                other_params[k] = v
        
        start_time = time.monotonic()
        res_data = await asyncfy_with_semaphore(lambda:self.client.chat.completions.create(
                            model=self.model,
                            temperature = temperature,
                            top_p = top_p,
                            messages=messages,**other_params))()
        
        if stream:            
            server = ray.get_actor("BLOCK_VLLM_STREAM_SERVER")
            request_id = [None]

            def writer(): 
                r = ""
                for response in res_data:                                        
                    v = response.choices[0].delta.content
                    r += v
                    request_id[0] = f"zhipu_{response.id}"
                    ray.get(server.add_item.remote(request_id[0], 
                                                    StreamOutputs(outputs=[SingleOutput(text=r,metadata=SingleOutputMeta(
                                                        input_tokens_count= -1,
                                                        generated_tokens_count= -1,
                                                    ))])
                                                    ))
                ray.get(server.mark_done.remote(request_id[0]))

            threading.Thread(target=writer,daemon=True).start()            
                               
            time_count= 10*100
            while request_id[0] is None and time_count > 0:
                await asyncio.sleep(0.01)
                time_count -= 1
            
            if request_id[0] is None:
                raise Exception("Failed to get request id")
            
            def write_running():
                return ray.get(server.add_item.remote(request_id[0], "RUNNING"))
                        
            await asyncio.to_thread(write_running)
            return [("",{"metadata":{"request_id":request_id[0],"stream_server":"BLOCK_VLLM_STREAM_SERVER"}})] 
      
        time_cost = time.monotonic() - start_time
        generated_text = res_data.choices[0].message.content        
        generated_tokens_count = res_data.usage.completion_tokens

        return [(generated_text,{"metadata":{
                        "request_id":res_data.id,
                        "input_tokens_count":res_data.usage.prompt_tokens,
                        "generated_tokens_count":generated_tokens_count,
                        "time_cost":time_cost,
                        "first_token_time":0,
                        "speed":float(generated_tokens_count)/time_cost,        
                    }})]





import ray
import json
import os
import uuid
import shutil
from typing import Optional, Tuple, List, Dict, Any
from datetime import datetime
from typing import Dict, Any,List,Generator
from pyjava.storage import streaming_tar as STar
from pyjava import RayContext
from pyjava.api.mlsql import DataServer
from byzerllm import BlockRow
from ray.air.util.torch_dist import (
    ActorHandle,
    _get_node_and_gpu_ids,
    _init_torch_distributed,
    get_address_and_port,
)
from . import qlora as QLoraTrainer
from byzerllm import restore_model
from .. import print_flush

DEFAULT_QLORA_CONFIG = {
    'output_dir': '',
    'model_name_or_path': '',
    'train_file': '',
    'num_train_epochs': 1,
    'per_device_train_batch_size': 1,
    'gradient_accumulation_steps': 16,
    'learning_rate': 0.0002,
    'max_seq_length': 1024,
    'logging_steps': 300,
    'save_steps': 500,
    'save_total_limit': 1,
    'lr_scheduler_type': 'cosine',
    'warmup_steps': 3000,
    'lora_rank': 64,
    'lora_alpha': 16,
    'lora_dropout': 0.05,
    'gradient_checkpointing': False,
    'disable_tqdm': False,
    'optim': 'paged_adamw_32bit',
    'seed': 42,
    'fp16': True,
    'report_to': 'tensorboard',
    'dataloader_num_workers': 0,
    'save_strategy': 'steps',
    'weight_decay': 0,
    'max_grad_norm': 0.3,
    'remove_unused_columns': False
 }

@ray.remote
class SFT:
    def __init__(self,data_refs:List[DataServer],sft_config:Dict[str,Any],train_params:Dict[str,str],sys_conf: Dict[str, str]) -> None:
        if "runIn" in sys_conf and sys_conf["runIn"] == "driver":
            raise Exception('''
                SFT can not run in driver. 
                Try the one of the following instructions:

                1. !byzerllm setup sft; 
                2. !byzerllm setup "runIn=executor"
            ''')
             
        self.sft_config = sft_config
        self.data_refs = data_refs
        self.train_params = train_params
        self.sys_conf = sys_conf

    def setup_tensorboard(self)->Optional[Tuple[str,int]]:        
        logging_dir = self.sft_config["logging_dir"]
        import subprocess               
        ip, port = get_address_and_port()
        log_dir = logging_dir            
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        tb_process = subprocess.Popen(['tensorboard', '--logdir', log_dir,"--port",str(port),"--host",ip], stdout=subprocess.PIPE, stderr=subprocess.PIPE)                                    
        self.tensorboard_pid = tb_process.pid
        return (ip,port)

    def train(self,args:List[str]):
        
        sft_name = self.train_params["name"] if "name" in self.train_params else f"sft-{self.sys_conf['OWNER']}"
        print_flush(f"[{sft_name}] SFT Config: {self.sft_config}")

        if not os.path.exists(self.sft_config["output_dir"]):
            os.makedirs(self.sft_config["output_dir"])
        
        train_file = self.sft_config["train_file"]
        if not os.path.exists(train_file): 
            os.makedirs(os.path.dirname(train_file))

        if "localModelDir" not in self.train_params:
            restore_model(self.conf,self.sft_config["model_name_or_path"])                                  
        
        
        # prepare data
        if self.data_refs:
            with open(train_file,"w") as f: 
                count = 0
                for item in RayContext.collect_from(self.data_refs):                
                    if "conversation" in item:
                        item["conversation"] = item["conversation"].tolist()
                        s = json.dumps(item,ensure_ascii=False)               
                        f.write(s+"\n")                    
                    elif "instruction" in item and "output" in item :
                        # support alpaca format data
                        history = item.get("history",[]) 
                        
                        if hasattr(history,"tolist"):
                            history = history.tolist()

                        input = item.get("input","")
                        conversation = [sub.tolist() for sub in history]
                        conversation = [{"human":x[0],"assistant":x[1]} for x in conversation]
                        latest_conversation = [{"human":item["instruction"]+"\n"+input,"assistant":item["output"]}] if "instruction" in item and item["instruction"] else []
                        s = json.dumps({
                            "category":"",
                            "conversation":conversation + latest_conversation,
                            "conversation_id":count,
                            "dataset":"",                
                        },ensure_ascii=False)               
                        f.write(s+"\n") 
                    else:
                        raise Exception(f"Unknown data format: {item}")                                            
                    count += 1       
        
        ip,port = self.setup_tensorboard()
        print_flush(f"[{sft_name}] Tensorboard is running at: {ip}:{port}")

        final_path = QLoraTrainer.train(json.dumps(self.sft_config,ensure_ascii=False), args, {
            "model_type": self.train_params.get("model_type","casual_lm"),"sft_name":sft_name
        })
        # copy the pretrained model to output dir
        if self.train_params.get("skipCopyPretrainedModel","false") == "false":
            print_flush(f'[{sft_name}] Copy pretrained model: {self.sft_config["model_name_or_path"]} to {os.path.join(final_path,"pretrained_model")}')        
            shutil.copytree(self.sft_config["model_name_or_path"],os.path.join(final_path,"pretrained_model"))
        
        # if detached, do not transfer the model to delta lake
        detached = self.train_params.get("detached","false") == "true"
        if detached:
            print_flush(f'''
              [{sft_name}] Train Actor is already finished. You can check the model in: {final_path}              
              ''') 
            return ([],0)
        
        # push the model to ray object store
        result = []
        count = 0
        print_flush(f"[{sft_name}] Store model({final_path}) to Ray object store")
        for item in STar.build_rows_from_file(final_path):
            if count % 1000 == 0:
                print_flush(f"[{sft_name}] Progress: {count} processed")
            count += 1    
            result.append(ray.put(item))
        
        print_flush(f'''
              [{sft_name}] Train Actor already finished.
              [{sft_name}] It may take a while to transfer the model from Ray object store to delta lake. 
              [{sft_name}] Try to check the progress in Byzer console or Byzer Notebook. 
              ''')    
        return (result,count) 

def sft_train(data_refs:List[DataServer],train_params:Dict[str,str],sys_conf: Dict[str, str])->Generator[BlockRow,Any,Any]:
    
    localPathPrefix = train_params.get("localPathPrefix","/tmp/byzerllm")
    
    current_time = datetime.now()
    formatted_time = current_time.strftime("%Y%m%d-%H-%M-%S")
    sft_name = train_params["name"] if "name" in train_params else f"sft-{sys_conf['OWNER']}-{formatted_time}"        
    
    rd = f"{sft_name}-{str(uuid.uuid4())}"    
    
    model_dir = os.path.join(localPathPrefix,rd,"pretrained_model")

    if "localModelDir" in train_params:
        model_dir = train_params["localModelDir"]

    output_dir = os.path.join(localPathPrefix,rd,"finetune_model")
    logging_dir = os.path.join(localPathPrefix,rd,"logging")
    data_dir = os.path.join(localPathPrefix,rd,"finetune_data")
    
    if "data_dir" in train_params:
        data_dir = train_params["data_dir"]

    data_file = os.path.join(data_dir,"data.jsonl")

    train_worker_conf = {}
    if "num_cpus" in sys_conf:
        train_worker_conf["num_cpus"] = float(sys_conf["num_cpus"])

    if "num_gpus" in sys_conf:
        train_worker_conf["num_gpus"] = float(sys_conf["num_gpus"])

    custom_resources = [(key.split("resource.")[1], float(sys_conf[key])) for key in
                        sys_conf.keys() if
                        key.startswith("resource.")]

    if len(custom_resources) > 0:
        train_worker_conf["resources"] = dict(custom_resources)   
    
    train_params_sft = {}
    
    for k,v in train_params.items():
        if k.startswith("sft."):
            # sft.float.num_train_epochs
            tpe = k.split(".")[1]
            new_k = k.split(".")[2]
            new_v = v
            if tpe == "float":
              new_v = float(v)
            elif tpe == "int":
                new_v = int(v)
            elif tpe == "bool":
                new_v = v == "true"
            elif tpe == "str":
                new_v = v
            elif tpe == "list":
                new_v = json.loads(v)
            elif tpe == "dict":
                new_v = json.loads(v)            
            train_params_sft[new_k] = new_v

    if "config" in train_params:
        train_params_sft = {**json.loads(train_params["config"]),**train_params_sft}

    sft_config = {
       **DEFAULT_QLORA_CONFIG,
       **train_params_sft,
       **{
           "output_dir":output_dir,
           "logging_dir": logging_dir,
           "model_name_or_path":model_dir,
           "train_file":data_file,
       }
    }         
    
    detached = train_params.get("detached","true") == "true"
    
    if detached:
        print_flush(f"[{sft_name}] Detached mode is enabled. ")
        train_actor = SFT.options(name=sft_name,lifetime="detached", **train_worker_conf).remote(data_refs,sft_config,train_params,sys_conf)
        train_actor.train.remote([])
        return [] 

    train_actor = SFT.options(name=sft_name,**train_worker_conf).remote(data_refs,sft_config,train_params,sys_conf)
    try:        
        items,obj_count = ray.get(train_actor.train.remote([]))
    except Exception as e:
        ray.kill(train_actor)
        raise e  
            
    print_flush(f"[{sft_name}] Transform Model from Ray object store to new storage(delta lake), total refs: {obj_count}. ")
    count = 0
    for item in items:
        if count % 1000 == 0:
            print_flush(f"[{sft_name}] Process: {float(count)/obj_count*100}%")
        count += 1
        yield ray.get(item)
        


from .conversable_agent import ConversableAgent
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from ...utils.client import ByzerLLM
from byzerllm.utils.retrieval import ByzerRetrieval
from .agent import Agent
from ray.util.client.common import ClientActorHandle, ClientObjectRef
import time
from . import get_agent_name,run_agent_func,ChatResponse


class CommonAgent(ConversableAgent):    

    DEFAULT_SYSTEM_MESSAGE = """You are a helpful AI assistant.
Based on the conversation and try you best to answer the user's quesion.
"""

    def __init__(
        self,
        name: str,
        llm: ByzerLLM,
        retrieval: ByzerRetrieval,
        chat_name:str,
        owner:str,
        code_agent: Union[Agent, ClientActorHandle,str],
        system_message: Optional[str] = DEFAULT_SYSTEM_MESSAGE,        
        is_termination_msg: Optional[Callable[[Dict], bool]] = None,
        max_consecutive_auto_reply: Optional[int] = None,
        human_input_mode: Optional[str] = "NEVER",
        code_execution_config: Optional[Union[Dict, bool]] = False,
        **kwargs,
    ):
        super().__init__(
            name,
            llm,retrieval,
            system_message,
            is_termination_msg,
            max_consecutive_auto_reply,
            human_input_mode,
            code_execution_config=code_execution_config,            
            **kwargs,
        )
        self.code_agent = code_agent
        self._reply_func_list = []
        self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.generate_llm_reply)           
        self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.check_termination_and_human_reply) 
        
        


import ray
import os
import time
import types
import copy
import asyncio
from typing import Any, Any, Dict, List, Tuple, Generator, Optional, Union
from byzerllm.utils.metrics import Metric
from byzerllm.utils import (
    VLLMStreamServer,
    StreamOutputs,
    SingleOutput,
    SingleOutputMeta,
    compute_max_new_tokens,
    tokenize_stopping_sequences,
)
from byzerllm.utils.tokenizer import (
    get_real_tokenizer,
    get_local_tokenizer,
    validate_args_engine_use_ray,
)
from byzerllm.utils.ray_utils import get_actor_info


try:
    from vllm.engine.async_llm_engine import (
        AsyncLLMEngine,
        AsyncEngineArgs,
        _AsyncLLMEngine,
    )
    from vllm import SamplingParams
    from vllm.utils import random_uuid
    from vllm.lora.request import LoRARequest
except ImportError:
    print(
        "vllm is not installed, if you want to use vllm backend,please install it by `pip install vllm`",
        flush=True,
    )
    pass
import torch
from transformers import (
    AutoTokenizer,
    AutoModelForCausalLM,
    BitsAndBytesConfig,
    StoppingCriteriaList,
    GenerationConfig,
)
from byzerllm.utils.types import StopSequencesCriteria
from ray.util.client.common import ClientActorHandle


INFERENCE_NAME = "auto"
INFER_TOKEN_METRICS = Metric()


def get_bool(params: Dict[str, str], key: str, default: bool = False) -> bool:
    if key in params:
        if isinstance(params[key], bool):
            return params[key]
        else:
            return params[key] == "true" or params[key] == "True"
    return default


def get_int(params: Dict[str, str], key: str, default: int = 0) -> int:
    if key in params:
        return int(params[key])
    return default


def get_float(params: Dict[str, str], key: str, default: float = 0.0) -> float:
    if key in params:
        return float(params[key])
    return default


def get_str(params: Dict[str, str], key: str, default: str = "") -> str:
    if key in params:
        return params[key]
    return default


async def async_get_meta(model):
    model: AsyncLLMEngine = model
    config = await model.get_model_config()
    tokenizer = model.local_tokenizer
    final_tokenizer = get_real_tokenizer(tokenizer)

    support_chat_template = (
        hasattr(final_tokenizer, "apply_chat_template")
        and hasattr(final_tokenizer, "chat_template")
        and final_tokenizer.chat_template is not None
    )

    meta = {
        "model_deploy_type": "proprietary",
        "backend": "ray/vllm",
        "support_stream": True,
        "support_chat_template": support_chat_template,
        "max_model_len": config.max_model_len,
        "architectures": getattr(config.hf_config, "architectures", []),
    }

    if not isinstance(model.engine, _AsyncLLMEngine):
        try:
            state = get_actor_info(model.engine)
            meta["engien_state"] = state.state
            meta["engine_actor_id"] = state.actor_id
            #  meta["engine_placement_group_id"] = model.placement_group.id.hex()
        except Exception as e:
            print(f"get engine state error:{e}", flush=True)

    return [meta]


async def async_vllm_chat(
    model,
    tokenizer,
    ins: str,
    his: List[Tuple[str, str]] = [],
    max_length: int = 4096,
    top_p: float = 0.95,
    temperature: float = 0.1,
    **kwargs,
):

    if "abort" in kwargs and "request_id" in kwargs:
        abort = get_bool(kwargs, "abort", False)
        request_id = kwargs["request_id"]
        if abort:
            await model.abort(request_id)
        return [
            (
                "",
                {
                    "metadata": {
                        "request_id": request_id,
                    }
                },
            )
        ]

    stream = get_bool(kwargs, "stream", False)
    request_id = kwargs["request_id"] if "request_id" in kwargs else random_uuid()
    n: int = 1
    best_of: Optional[int] = get_int(kwargs, "best_of", None)
    presence_penalty: float = float(kwargs.get("presence_penalty", 0.0))
    frequency_penalty: float = float(kwargs.get("frequency_penalty", 0.0))
    top_k: int = int(kwargs.get("top_k", -1))
    use_beam_search: bool = get_bool(kwargs, "use_beam_search", False)
    stop: Union[None, str, List[str]] = kwargs["stop"] if "stop" in kwargs else None
    ignore_eos: bool = get_bool(kwargs, "ignore_eos", False)
    max_tokens: int = max_length
    logprobs: Optional[int] = get_int(kwargs, "logprobs", None)

    adapter_name_or_path: Optional[str] = kwargs.get("adapter_name_or_path", None)
    lora_name: Optional[str] = get_str(kwargs, "lora_name", "default")
    lora_int_id: Optional[str] = get_int(kwargs, "lora_int_id", 1)

    lora_request = None
    if adapter_name_or_path is not None:
        lora_request = LoRARequest(
            lora_name=lora_name, lora_int_id=lora_int_id, lora_local_path=adapter_name_or_path
        )

    # repetition_penalty: float = float(kwargs.get("repetition_penalty",1.1))

    other_params = {}
    if "early_stopping" in kwargs:
        other_params["early_stopping"] = bool(kwargs["early_stopping"])

    if "repetition_penalty" in kwargs:
        other_params["repetition_penalty"] = float(kwargs["repetition_penalty"])

    if "stop_token_ids" in kwargs:
        stop_token_ids = kwargs["stop_token_ids"]
        if isinstance(stop_token_ids, str):
            stop_token_ids = [int(i) for i in stop_token_ids.split(",")]
        else:
            stop_token_ids = kwargs["stop_token_ids"]
        other_params["stop_token_ids"] = stop_token_ids

    sampling_params = SamplingParams(
        temperature=temperature,
        n=n,
        best_of=best_of,
        presence_penalty=presence_penalty,
        frequency_penalty=frequency_penalty,
        top_k=top_k,
        use_beam_search=use_beam_search,
        stop=stop,
        ignore_eos=ignore_eos,
        logprobs=logprobs,
        top_p=top_p,
        max_tokens=max_tokens,
        **other_params,
    )

    current_time_milliseconds = int(time.time() * 1000)

    if stream:
        server = ray.get_actor("VLLM_STREAM_SERVER")

        async def writer():
            results_generator = model.generate(
                ins, sampling_params, request_id, lora_request=lora_request
            )
            async for request_output in results_generator:
                v = StreamOutputs(
                    outputs=[
                        SingleOutput(
                            text=item.text,
                            metadata=SingleOutputMeta(
                                input_tokens_count=len(request_output.prompt_token_ids),
                                generated_tokens_count=len(item.token_ids),
                            ),
                        )
                        for item in request_output.outputs
                    ]
                )
                await server.add_item.remote(request_output.request_id, v)
            # mark the request is done
            await server.mark_done.remote(request_output.request_id)

        asyncio.create_task(writer())
        await server.add_item.remote(request_id, "RUNNING")
        return [
            (
                "",
                {
                    "metadata": {
                        "request_id": request_id,
                        "stream_server": "VLLM_STREAM_SERVER",
                    }
                },
            )
        ]

    results_generator = model.generate(ins, sampling_params, request_id,lora_request=lora_request)
    final_output = None
    first_token_time = current_time_milliseconds
    async for request_output in results_generator:
        if (
            first_token_time == current_time_milliseconds
            and request_output.outputs
            and len(request_output.outputs[0].token_ids) > 0
        ):
            first_token_time = int(time.time() * 1000)
        final_output = request_output
    assert final_output is not None

    text_outputs = [output for output in final_output.outputs]
    generated_text = text_outputs[0].text
    prob = text_outputs[0].cumulative_logprob

    current_time_milliseconds2 = int(time.time() * 1000)

    input_tokens_count = len(final_output.prompt_token_ids)
    generated_tokens_count = len(text_outputs[0].token_ids)
    time_cost = current_time_milliseconds2 - current_time_milliseconds
    print(
        f"cost: {time_cost}ms first_token:{first_token_time-current_time_milliseconds}ms speed: {float(generated_tokens_count)/time_cost*1000} tokens/s total_tokens_count:{input_tokens_count + generated_tokens_count} request_id:{final_output.request_id}  input_tokens_count:{input_tokens_count} generated_tokens_count:{generated_tokens_count}",
        flush=True,
    )

    INFER_TOKEN_METRICS.inc(
        f"infer_{INFERENCE_NAME}_input_tokens_num",
        input_tokens_count,
        tags={"request_id": final_output.request_id},
    )
    INFER_TOKEN_METRICS.inc(
        f"infer_{INFERENCE_NAME}_output_tokens_num",
        generated_tokens_count,
        tags={"request_id": final_output.request_id},
    )
    INFER_TOKEN_METRICS.push()

    return [
        (
            generated_text,
            {
                "metadata": {
                    "request_id": final_output.request_id,
                    "input_tokens_count": input_tokens_count,
                    "generated_tokens_count": generated_tokens_count,
                    "time_cost": time_cost,
                    "first_token_time": first_token_time - current_time_milliseconds,
                    "speed": float(generated_tokens_count) / time_cost * 1000,
                    "prob": prob,
                }
            },
        )
    ]


def from_vllm_backend(
    model_path, infer_params: Dict[str, str] = {}, sys_conf: Dict[str, str] = {}
):
    infer_mode = sys_conf.get("infer_backend", "transformers")
    num_gpus = int(sys_conf.get("num_gpus", 1))
    print(f"infer_mode:{infer_mode} tensor_parallel_size: {num_gpus}")
    global INFERENCE_NAME
    INFERENCE_NAME = infer_params.get("udfName", "auto")

    try:
        ray.get_actor("VLLM_STREAM_SERVER")
    except ValueError:
        try:
            ray.remote(VLLMStreamServer).options(
                name="VLLM_STREAM_SERVER", lifetime="detached", max_concurrency=1000
            ).remote()
        except Exception as e:
            print(f"create VLLM_STREAM_SERVER error:{e}", flush=True)

    worker_use_ray: bool = get_bool(infer_params, "backend.worker_use_ray", True)

    engine_use_ray: bool = validate_args_engine_use_ray()
    if "backend.engine_use_ray" in infer_params:
        engine_use_ray = get_bool(infer_params, "backend.engine_use_ray", False)

    tensor_parallel_size: int = num_gpus
    gpu_memory_utilization: float = float(
        infer_params.get("backend.gpu_memory_utilization", 0.90)
    )
    disable_log_stats: bool = get_bool(infer_params, "backend.disable_log_stats", False)

    other_params = {}

    def parse_value(v):
        if not isinstance(v, str):
            return v
        v = v.strip()
        if v.lower() in ['true', 'false']:
            return v.lower() == 'true'
        try:
            return int(v)
        except ValueError:
            try:
                return float(v)
            except ValueError:
                return v

    for k, v in infer_params.items():
        if k.startswith("backend.") and k not in [
            "backend.worker_use_ray",
            "backend.engine_use_ray",
            "backend.gpu_memory_utilization",
            "backend.disable_log_stats",
        ]:
            new_k = k[len("backend."):]
            other_params[new_k] = parse_value(v)

    engine_args = AsyncEngineArgs(
        engine_use_ray=engine_use_ray,
        model=model_path,
        worker_use_ray=worker_use_ray,
        tensor_parallel_size=tensor_parallel_size,
        gpu_memory_utilization=gpu_memory_utilization,
        disable_log_stats=disable_log_stats,
        **other_params,
    )
    llm = AsyncLLMEngine.from_engine_args(engine_args)
    tokenizer = get_local_tokenizer(llm, engine_args)
    llm.local_tokenizer = tokenizer
    llm.async_stream_chat = types.MethodType(async_vllm_chat, llm)
    llm.async_get_meta = types.MethodType(async_get_meta, llm)
    return (llm, tokenizer)



import pytest
import byzerllm

@pytest.fixture(scope="module")
def setup():
    byzerllm.connect_cluster()
    llm = byzerllm.ByzerLLM()
    model = "qianwen_chat"
    llm.setup_default_model_name(model)
    llm.setup_template(model,"auto")
    return llm

class Wow:
    def __init__(self,llm):
        self.llm = llm

    @byzerllm.prompt(lambda self:self.llm)
    def test(self,s:str)->str: 
        '''
        Hello, {{ s }}!
        '''

    @byzerllm.prompt()
    def test2(self,s:str)->str: 
        '''
        Hello, {{ s }}!
        '''    

@byzerllm.prompt(options={"llm_config":{}})
def test(s:str)->str: 
    '''
    Hello, {{ s }}!
    '''  

def test_prompt_class_method(setup):
    w = Wow(setup)
    assert w.test.prompt("你是谁") == "\nHello, 你是谁!"
    assert isinstance(w.test.run("你是谁"),str)

def test_prompt_class_method_no_llm(setup):    
    w = Wow(setup)
    assert w.test2.prompt("你是谁") == "\nHello, 你是谁!"
    
def test_prompt_standalone_func(setup):
    test.with_llm(setup)
    assert test.prompt("你是谁") == "\nHello, 你是谁!"    
    
def test_prompt_options(setup):
    w = Wow(setup)    
    result = w.test.options({"model":"kimi_8k_chat","llm_config":{"max_length":1000}}).run("你是谁")
    assert "MoonshotAI" in result
    
def test_prompt_with_llm(setup):
    w = Wow(setup)
    assert isinstance(w.test.with_llm(setup).run("你是谁"))


from byzerllm.utils.client import ByzerLLM
from byzerllm.utils.retrieval import ByzerRetrieval
from byzerllm.apps.llama_index.byzerai import ByzerAI
from byzerllm.apps.llama_index.byzerai_embedding import ByzerAIEmbedding
from byzerllm.apps.llama_index.byzerai_docstore import ByzerAIDocumentStore
from byzerllm.apps.llama_index.byzerai_index_store import ByzerAIIndexStore
from byzerllm.apps.llama_index.byzerai_vectordb import ByzerAIVectorStore
from llama_index.core.service_context import ServiceContext
from llama_index.core.storage import StorageContext
from typing import Optional

def get_service_context(llm:ByzerLLM,**kargs):        
    service_context = ServiceContext.from_defaults(llm=ByzerAI(llm=llm),embed_model=ByzerAIEmbedding(llm=llm),**kargs)
    return service_context

def get_storage_context(llm:ByzerLLM,retrieval:ByzerRetrieval,
                        chunk_collection:Optional[str]="default",
                        namespace:Optional[str]=None,                        
                        **kargs):
    vector_store = ByzerAIVectorStore(llm=llm, retrieval=retrieval,chunk_collection=chunk_collection)
    docstore = ByzerAIDocumentStore(llm=llm, retrieval=retrieval,namespace=namespace)
    index_store = ByzerAIIndexStore(llm=llm, retrieval=retrieval,namespace=namespace)
    storage_context = StorageContext.from_defaults(
        docstore=docstore,
        vector_store=vector_store,
        index_store=index_store,
        **kargs
    )
    return storage_context    



"""Langchain Embedding Wrapper Module."""

from typing import TYPE_CHECKING, List, Optional
from llama_index.core.base.embeddings.base import (
    DEFAULT_EMBED_BATCH_SIZE,
    BaseEmbedding,
)

try:
    from llama_index.core.bridge.pydantic import Field, PrivateAttr
except ImportError:
    from pydantic import Field, PrivateAttr

from byzerllm.utils.client import ByzerLLM
from byzerllm.utils.langutil import asyncfy_with_semaphore


class ByzerAIEmbedding(BaseEmbedding):

    _llm: ByzerLLM = PrivateAttr()

    def __init__(
        self,
        llm: ByzerLLM,
    ):
        self._llm = llm
        super().__init__(
            model_name=self._llm.default_emb_model_name,
        )

    @classmethod
    def class_name(cls) -> str:
        return "ByzerAIEmbedding"

    def _get_query_embedding(self, query: str) -> List[float]:
        """Get query embedding."""               
        return self._llm.emb_query(query)[0].output[0:1024]

    async def _aget_query_embedding(self, query: str) -> List[float]:
        return asyncfy_with_semaphore(self._get_query_embedding, query)

    async def _aget_text_embedding(self, text: str) -> List[float]:
        return asyncfy_with_semaphore(self._get_query_embedding, text)

    def _get_text_embedding(self, text: str) -> List[float]:
        return self._get_query_embedding(text)

    def _get_text_embeddings(self, texts: List[str]) -> List[List[float]]:
        """Get text embeddings."""
        return [self._get_text_embedding(text) for text in texts]



from . import Message, MessageStore
import time

class MemoryStore(MessageStore):
    def __init__(self):
        self.messages = {}

    def put(self, message: Message):
        if message.id not in self.messages:
            self.messages[message.id] = []
        v = self.messages[message.id]
        v.append(message)
        if message.m is None:
            return self
    
        # clean up old messages
        target = -1 
        for idx,item in enumerate(v):
            if time.monotonic() - item.timestamp > 24*60*60:
                target = idx
                break
        if target >= 0:
            v = v[target:]
            self.messages[message.id] = v
        
        # clean up message groups
        remove_keys = []
        for key in list(self.messages.keys()):
            if time.monotonic() -  self.messages[key][-1].timestamp  > 24*60*60:
                remove_keys.append(key)

        for key in remove_keys:
            del self.messages[key]        

        return self
    
    def get(self, id: str):
        v = self.messages.get(id,[])
        if len(v)>0 and v[-1].m is None:
            del self.messages[id]
            return v
        return v
    
    def clear(self, id: str):
        if id in self.messages:
            del self.messages[id]
        
    


import sys
import subprocess
import os
import platform
import subprocess
from dataclasses import dataclass
import urllib.request
import tarfile
import zipfile
from loguru import logger
from packaging import version
import re
from os.path import expanduser

def is_apple_m_series():
    # 检查系统是否是 macOS
    if platform.system() != 'Darwin':
        return False

    try:
        # 使用 subprocess 获取系统信息
        result = subprocess.run(['sysctl', 'machdep.cpu.brand_string'], capture_output=True, text=True)
        # 检查输出中是否包含 "Apple M"
        if 'Apple M' in result.stdout:
            return True
    except Exception as e:
        print(f"An error occurred: {e}")
    
    return False

@dataclass
class EnvInfo:
   os_name: str
   os_version: str
   python_version: str
   conda_env: str
   virtualenv: str
   has_bash: bool
   java_home: str
   java_version: str
   cpu: str

def find_jdk21_from_dir(directory):
    
    entries = os.listdir(directory)
    
    # Iterate through the entries and filter directories that match the pattern
    for entry in entries:        
        match = "jdk-21" in entry.lower()
        if match and os.path.isdir(os.path.join(directory, entry)):
            return os.path.join(directory, entry)  # Return the directory name if found
    
    return None  # Return None if no matching directory is found     

def get_latest_byzer_retrieval_lib(directory):
    # Define the regex pattern for matching the directories
    pattern = r'^byzer-retrieval-lib-(\d+\.\d+\.\d+)$'
    
    # List all entries in the directory
    entries = os.listdir(directory)
    
    # Initialize an empty list to hold (version, directory name) tuples
    versions = []
    
    # Iterate through the entries and filter directories that match the pattern
    for entry in entries:
        match = re.match(pattern, entry)
        if match:
            # Extract the version part from the directory name
            ver = match.group(1)
            # Append the version and directory name as a tuple
            versions.append((version.parse(ver), entry))
    
    # Sort the list of tuples by the version (first element of the tuple)
    versions.sort()
    
    # Get the last element from the sorted list (the one with the highest version)
    if versions:
        return versions[-1][1]  # Return the directory name of the latest version
    else:
        return None  # Return None if no matching directory is found


def detect_env() -> EnvInfo:
    os_name = sys.platform
    os_version = ""
    cpu = "x86_64"
    if os_name == "win32":
        os_version = sys.getwindowsversion().major
    elif os_name == "darwin":
        os_version = subprocess.check_output(["sw_vers", "-productVersion"]).decode('utf-8').strip() 
        if is_apple_m_series():
            cpu = "arm64"
    elif os_name == "linux":
        os_version = subprocess.check_output(["uname", "-r"]).decode('utf-8').strip()
        if "aarch64" in subprocess.check_output(["uname", "-m"]).decode('utf-8').strip().lower():
            cpu = "arm64"

    python_version = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"

    conda_env = os.environ.get("CONDA_DEFAULT_ENV")

    virtualenv = os.environ.get("VIRTUAL_ENV")

    has_bash = True
    try:
        subprocess.check_output(["bash", "--version"])
    except:
        has_bash = False 

    default_path = os.path.join(expanduser("~"), ".auto-coder")
    if not os.path.exists(default_path):
        default_path = os.path.join(expanduser("~"), ".byzerllm")
        if not os.path.exists(default_path):
            default_path = ""
    
    java_home = os.environ.get("JAVA_HOME")
    java_version = "0"
    if default_path:
        jdk21_dir = find_jdk21_from_dir(default_path)        
        if jdk21_dir:
            java_home = jdk21_dir
            if os_name == "darwin":
                java_home = os.path.join(java_home, "Contents", "Home")                
            java_version = "21"
    else:                
        if java_home:
            try:
                output = subprocess.check_output([f"{java_home}/bin/java", "-version"], stderr=subprocess.STDOUT, universal_newlines=True)
                version_line = output.splitlines()[0]
                version_match = re.search(r'version "(\d+)', version_line)
                if version_match:
                    java_version = version_match.group(1)
            except (subprocess.CalledProcessError, ValueError) as e:
                logger.warning(f"Error checking Java version: {str(e)}")

    return EnvInfo(
        os_name=os_name,
        os_version=os_version,
        python_version=python_version,
        conda_env=conda_env,
        virtualenv=virtualenv,
        has_bash=has_bash,
        java_version=java_version,
        java_home=java_home,       
        cpu=cpu
    )


def download_with_progressbar(url, filename):
   def progress(count, block_size, total_size):
       percent = int(count * block_size * 100 / total_size)
       print(f"\rDownload progress: {percent}%", end="")

   urllib.request.urlretrieve(url, filename, reporthook=progress)


def download_and_install_jdk21(env_info: EnvInfo, install_dir: str):
   jdk_download_url = ""
   if env_info.os_name == "linux":
       if env_info.cpu == "arm64":
           jdk_download_url = "https://download.oracle.com/java/21/archive/jdk-21.0.2_linux-aarch64_bin.tar.gz"
       else:
           jdk_download_url = "https://download.java.net/java/GA/jdk21.0.2/f2283984656d49d69e91c558476027ac/13/GPL/openjdk-21.0.2_linux-x64_bin.tar.gz"
   elif env_info.os_name == "darwin":
       if env_info.cpu == "arm64": 
           jdk_download_url = "https://download.oracle.com/java/21/archive/jdk-21.0.2_macos-aarch64_bin.tar.gz"
       else:
           jdk_download_url = "https://download.java.net/java/GA/jdk21.0.2/f2283984656d49d69e91c558476027ac/13/GPL/openjdk-21.0.2_macos-x64_bin.tar.gz"
   elif env_info.os_name == "win32":
       jdk_download_url = "https://download.java.net/java/GA/jdk21.0.2/f2283984656d49d69e91c558476027ac/13/GPL/openjdk-21.0.2_windows-x64_bin.zip"

   if jdk_download_url:
       logger.info(f"Downloading JDK 21 from {jdk_download_url}")
       download_path = os.path.join(install_dir, os.path.basename(jdk_download_url))
       download_with_progressbar(jdk_download_url, download_path)

       if env_info.os_name == "win32":
           with zipfile.ZipFile(download_path, "r") as zip_ref:
               zip_ref.extractall(install_dir)
       else:
           with tarfile.open(download_path, "r:gz") as tar:
               tar.extractall(path=install_dir)

       logger.info("JDK 21 downloaded and installed successfully")
   else:
       logger.warning(f"No JDK 21 download URL found for {env_info.os_name}")



from typing import List,Dict,Annotated,Any,Optional
from byzerllm.utils.client import ByzerLLM,message_utils,code_utils
from byzerllm.utils.retrieval import ByzerRetrieval
import copy
import json
import pydantic
from . import QueryRewriteResult,Action

class QueryCondition:
    '''
    this tool is used to extract the key messages from the user's question which
    is used to generate the sql code conditions e.g. filter conditions, group by conditions, order by conditions
    or aggregate fields.
    '''
    def __init__(self,llm:ByzerLLM, retrieval:ByzerRetrieval,
                 sys_message:List[Dict[str,Any]],
                 messages:List[Dict[str,Any]],
                 **kwargs):
        self.messages = messages
        self._system_message = sys_message
        self.llm = llm
        self.retrieval = retrieval
        self.params = kwargs        

    def apply(self)->QueryRewriteResult:        
        m = copy.deepcopy(self.messages[-1])
        time_msg = self.params.get("time_msg","")                       
        class KV(pydantic.BaseModel):
            name:str = pydantic.Field(description="表字段名称")
            value:Optional[str] = pydantic.Field(default="",description="从问题得到的值")
            tpe:Optional[str] = pydantic.Field(default="",description="类型，过滤条件，指标，分组条件等")
                                       

        class Conditions(pydantic.BaseModel):
            items:List[KV]=pydantic.Field(...,description="the key value pairs of the query conditions")

        temp_conversation = [{"role":"user","content":'''
首先根据我的问题，关联前面的对话，尤其是前面的表结构表结构信息，示例数据，表统计信息等，找到我当前问题中的关键信息,
诸如过滤条件，指标，分组条件。不需要生成SQL。                      

具体请按如下方式步骤补充信息，务必一步一步来：

1. 回顾前面的会话，对提到的表结构信息，示例数据，表统计信息等进行回顾，列出字段列表。
2. 对当前问题进行拆解，找到可能的过滤条件，指标，分组条件等。以 字段名称=值 的形式罗列出来。
3. 根据表统计信息中列表，对第二步过滤条件的值进行修正。具体做法是，如果过滤条件字段在表统计信息中有枚举值，
检查过滤字段的值是否在枚举值中，如果不在，找到最接近的枚举值，修正过滤条件的值。
4. 对最后修正的结果，重新以 Json 格式进行输出

请输出每一步的结果。                                                                                          
        '''}] 

         
        t = self.llm.chat_oai(conversations=message_utils.padding_messages_merge(
            self._system_message  + self.messages + self.params.get("temp_conversation",temp_conversation)),
            response_class=Conditions,
            enable_default_sys_message=True)
        action = Action.CONTINUE
        key_msg = ""
        
        if t[0].value:
            v:Conditions = t[0].value
            for item in v.items:
                key_msg += f''' {item.name}={item.value}'''

            old_content = m["content"]
            m["content"] = f'''补充信息：{time_msg} {key_msg} \n原始问题：{old_content} '''
            print(f'final query:{m["content"]}\n\n',flush=True)                                              
        
        return QueryRewriteResult(message = m,action = action,extra_info={"key_msg":key_msg})        
             


    


from transformers import AutoTokenizer, AutoModel
import transformers
import torch
import os
from typing import Any,Any,Dict, List,Tuple,Generator


from pyjava.api.mlsql import DataServer
from .. import BlockRow
from .. import parse_params

  
def get_meta(self): 
    config = self.config   
    return [{
        "model_deploy_type": "proprietary",
        "backend":"transformers",
        "max_model_len":getattr(config, "model_max_length", -1),
        "architectures":getattr(config, "architectures", [])
    }]

def stream_chat(self,tokenizer,ins:str, his:List[Tuple[str,str]]=[],  
        max_length:int=4096, 
        top_p:float=0.95,
        temperature:float=0.1,**kwargs):
        
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    tokens = tokenizer(ins, return_token_type_ids=False,return_tensors="pt").to(device)
    response = self.generate(
        input_ids=tokens["input_ids"],
        max_new_tokens=max_length,
        repetition_penalty=1.05,
        temperature=temperature,
        eos_token_id=tokenizer.eos_token_id
    )
    answer = tokenizer.decode(response[0][tokens["input_ids"].shape[1]:], skip_special_tokens=True)
    return [(answer,"")]


def init_model(model_dir,infer_params:Dict[str,str]={},sys_conf:Dict[str,str]={}): 
    pretrained_model_dir = os.path.join(model_dir,"pretrained_model")
    adaptor_model_dir = model_dir
    is_adaptor_model = os.path.exists(pretrained_model_dir)
    
    if not is_adaptor_model:        
        pretrained_model_dir = model_dir

    params = parse_params(infer_params,"infer")
    load_in_4bit = params.get("load_in_4bit",False)
    
    tokenizer = AutoTokenizer.from_pretrained(pretrained_model_dir,trust_remote_code=True)    
    model = AutoModel.from_pretrained(pretrained_model_dir,trust_remote_code=True ).half().cuda()
    if is_adaptor_model:
        from peft import PeftModel
        model = PeftModel.from_pretrained(model, adaptor_model_dir)
        
    model.eval()       
    import types
    # model.stream_chat = types.MethodType(stream_chat, model)  
    model.get_meta = types.MethodType(get_meta, model)     
    return (model,tokenizer)


def sft_train(data_refs:List[DataServer],
              train_params:Dict[str,str],
              conf: Dict[str, str])->Generator[BlockRow,Any,Any]:
    from ..utils.sft import sft_train as common_sft_train
    return common_sft_train(data_refs,train_params,conf) 






import os
import time
import ray
import asyncio
from typing import List, Dict, Any,Union
import traceback
import uuid
import threading
import json
import inspect

from llama_cpp import Llama
import llama_cpp
from byzerllm.utils.types import ( 
    BlockVLLMStreamServer,   
    StreamOutputs,
    SingleOutput,
    SingleOutputMeta,    
)


def get_bool(params:Dict[str,str],key:str,default:bool=False)->bool:
    if key in params:
        if isinstance(params[key],bool):
            return params[key]
        else:            
            return params[key] == "true" or params[key] == "True"
    return default


def get_int(params:Dict[str,str],key:str,default:int=0)->int:
    if key in params:
        return int(params[key])
    return default

def get_float(params:Dict[str,str],key:str,default:float=0.0)->float:
    if key in params:
        return float(params[key])
    return default

def get_str(params:Dict[str,str],key:str,default:str="")->str:
    if key in params:
        return params[key]
    return default  


def params_to_llama_kwargs(params: Dict[str, str]) -> Dict[str, Any]:
    llama_kwargs = {        
        "n_gpu_layers": get_int(params, "n_gpu_layers", 0),
        "split_mode": get_int(params, "split_mode", llama_cpp.LLAMA_SPLIT_MODE_LAYER),
        "main_gpu": get_int(params, "main_gpu", 0),
        "tensor_split": [float(x) for x in get_str(params, "tensor_split", "").split(",")] if "tensor_split" in params else None,
        "vocab_only": get_bool(params, "vocab_only", False),
        "use_mmap": get_bool(params, "use_mmap", True),
        "use_mlock": get_bool(params, "use_mlock", False),
        "seed": get_int(params, "seed", llama_cpp.LLAMA_DEFAULT_SEED),
        "n_ctx": get_int(params, "n_ctx", 512),
        "n_batch": get_int(params, "n_batch", 512),
        "n_threads": get_int(params, "n_threads") if "n_threads" in params else None,
        "n_threads_batch": get_int(params, "n_threads_batch") if "n_threads_batch" in params else None,
        "rope_scaling_type": get_int(params, "rope_scaling_type", llama_cpp.LLAMA_ROPE_SCALING_TYPE_UNSPECIFIED),
        "pooling_type": get_int(params, "pooling_type", llama_cpp.LLAMA_POOLING_TYPE_UNSPECIFIED),
        "rope_freq_base": get_float(params, "rope_freq_base", 0.0),
        "rope_freq_scale": get_float(params, "rope_freq_scale", 0.0),
        "yarn_ext_factor": get_float(params, "yarn_ext_factor", -1.0),
        "yarn_attn_factor": get_float(params, "yarn_attn_factor", 1.0),
        "yarn_beta_fast": get_float(params, "yarn_beta_fast", 32.0),
        "yarn_beta_slow": get_float(params, "yarn_beta_slow", 1.0),
        "yarn_orig_ctx": get_int(params, "yarn_orig_ctx", 0),
        "logits_all": get_bool(params, "logits_all", False),
        "embedding": get_bool(params, "embedding", False),
        "offload_kqv": get_bool(params, "offload_kqv", True),
        "last_n_tokens_size": get_int(params, "last_n_tokens_size", 64),
        "lora_base": get_str(params, "lora_base") if "lora_base" in params else None,
        "lora_scale": get_float(params, "lora_scale", 1.0),
        "lora_path": get_str(params, "lora_path") if "lora_path" in params else None,
        "numa": get_int(params, "numa") if "numa" in params else False,
        "chat_format": get_str(params, "chat_format") if "chat_format" in params else None,
        "verbose": get_bool(params, "verbose", True),
        "type_k": get_int(params, "type_k") if "type_k" in params else None,
        "type_v": get_int(params, "type_v") if "type_v" in params else None,
    }
    new_llama_kwargs = {}
    for k in list(params.keys()):
        new_llama_kwargs[k]=llama_kwargs[k]            
    return new_llama_kwargs

class LlamaCppBackend:

    def __init__(self,model_path, infer_params: Dict[str, str] = {}, sys_conf: Dict[str, str] = {}):
        targets = params_to_llama_kwargs(infer_params)
        self.model = Llama(model_path=model_path,**targets)        
        self.meta = {
            "model_deploy_type": "saas",
            "backend":"llama_cpp",
            "support_stream": True,            
        }

        try:
            ray.get_actor("BLOCK_VLLM_STREAM_SERVER") 
        except ValueError:  
            try:          
                ray.remote(BlockVLLMStreamServer).options(name="BLOCK_VLLM_STREAM_SERVER",lifetime="detached",max_concurrency=1000).remote()
            except Exception as e:
                print(f"create BlockVLLMStreamServer error:{e}",flush=True)                


    def get_meta(self):
        return [self.meta]

    def process_input(self, ins: Union[str, List[Dict[str, Any]]]):
        
        if isinstance(ins, list):
            return ins
        
        content = []
        try:
            ins_json = json.loads(ins)
        except:            
            return ins
        
        content = []
        for item in ins_json:
            if "image" in item or "image_url" in item:
                image_data = item.get("image",item.get("image_url",""))
                ## "data:image/jpeg;base64," 
                if not image_data.startswith("data:"):
                    image_data = "data:image/jpeg;base64," + image_data                                                                                
                content.append({"image_url": {"url":image_data},"type": "image_url",})
            elif "text" in item:
                text_data = item["text"]
                content.append({"text": text_data,"type":"text"})
        if not content:
            return ins
        
        return content   
        
    async def async_get_meta(self):
        return self.get_meta()

    async def async_stream_chat(self, tokenizer, ins: str, his: List[Dict[str, str]] = [],
                 max_length: int = 4090, top_p: float = 0.95, temperature: float = 0.1, **kwargs):
        return await self.generate(tokenizer=tokenizer, ins=ins, his=his, max_length=max_length, top_p=top_p, temperature=temperature, **kwargs)
    
    def embed_query(self, ins: str, **kwargs):                     
        resp = self.model.create_embedding(input = [ins])
        embedding = resp.data[0].embedding
        usage = resp.usage
        return (embedding,{"metadata":{
                "input_tokens_count":usage.prompt_tokens,
                "generated_tokens_count":0}})            
   
    async def generate(self, tokenizer, ins: str, his: List[Dict[str, str]] = [],
                 max_length: int = 4090, top_p: float = 0.95, temperature: float = 0.1, **kwargs):
                           
        messages = [{"role":message["role"],"content":self.process_input(message["content"])} for message in his] + [{"role": "user", "content": self.process_input(ins)}]

        stream = kwargs.get("stream",False)
        
        server = ray.get_actor("BLOCK_VLLM_STREAM_SERVER")
        request_id = [None]
        
        def writer():
            try:
                r = ""       
                response = self.model.create_chat_completion_openai_v1(
                                    messages=messages,                                    
                                    stream=True, 
                                    max_tokens=max_length,
                                    temperature=temperature,
                                    top_p=top_p                                                                        
                                )                                    
                request_id[0] = str(uuid.uuid4())                

                for chunk in response:                                                              
                    content = chunk.choices[0].delta.content or ""
                    r += content        
                    if hasattr(chunk,"usage"):
                        input_tokens_count = chunk.usage.prompt_tokens
                        generated_tokens_count = chunk.usage.completion_tokens
                    else:
                        input_tokens_count = 0
                        generated_tokens_count = 0
                    ray.get(server.add_item.remote(request_id[0], 
                                                    StreamOutputs(outputs=[SingleOutput(text=r,metadata=SingleOutputMeta(
                                                        input_tokens_count=input_tokens_count,
                                                        generated_tokens_count=generated_tokens_count,
                                                    ))])
                                                    ))                                                   
            except:
                traceback.print_exc()            
            ray.get(server.mark_done.remote(request_id[0]))

        if stream:
            threading.Thread(target=writer,daemon=True).start()            
                            
            time_count= 10*100
            while request_id[0] is None and time_count > 0:
                await asyncio.sleep(0.01)
                time_count -= 1
            
            if request_id[0] is None:
                raise Exception("Failed to get request id")
            
            def write_running():
                return ray.get(server.add_item.remote(request_id[0], "RUNNING"))
                        
            await asyncio.to_thread(write_running)
            return [("",{"metadata":{"request_id":request_id[0],"stream_server":"BLOCK_VLLM_STREAM_SERVER"}})]
        else:
            try:
                start_time = time.monotonic()
                response = self.model.create_chat_completion_openai_v1(
                                    messages=messages,                                    
                                    max_tokens=max_length,
                                    temperature=temperature,
                                    top_p=top_p                            
                                )

                generated_text = response.choices[0].message.content
                generated_tokens_count = response.usage.completion_tokens
                input_tokens_count = response.usage.prompt_tokens
                time_cost = time.monotonic() - start_time
                return [(generated_text,{"metadata":{
                            "request_id":response.id,
                            "input_tokens_count":input_tokens_count,
                            "generated_tokens_count":generated_tokens_count,
                            "time_cost":time_cost,
                            "first_token_time":0,
                            "speed":float(generated_tokens_count)/time_cost,        
                        }})]
            except Exception as e:
                print(f"Error: {e}")
                raise e


import time
from typing import List, Tuple, Dict, Any, Union
import base64
import json
import librosa
import io
from transformers import Qwen2AudioForConditionalGeneration, AutoProcessor

def get_meta(self):
    return [
        {
            "model_deploy_type": "proprietary",
            "backend": "transformers",
            "message_format": True
        }
    ]

def process_input(ins: Union[str, List[Dict[str, Any]], Dict[str, Any]]):
    if isinstance(ins, list) or isinstance(ins, dict):
        return ins

    content = []
    try:
        ins_json = json.loads(ins)
    except:
        return ins

    if isinstance(ins_json, dict):
        return ins_json

    for item in ins_json:
        if ("audio" in item or "audio_url" in item) and "type" not in item:
            audio_data = item.get("audio", item.get("audio_url", ""))
            if not audio_data.startswith("data:"):
                audio_data = "data:audio/wav;base64," + audio_data

            # other_fields = {k: v for k, v in item.items() if k not in ["audio", "text", "audio_url"]}
            content.append(
                {
                    "audio_url": audio_data,
                    "type": "audio",
                }
            )

        if "text" in item and "type" not in item:
            text_data = item["text"]
            content.append({"text": text_data, "type": "text"})

        if "type" in item and item["type"] in ["text", "audio"]:
            content.append(item)

    if not content:
        return ins

    return content

def stream_chat(
    self,
    tokenizer,
    ins: str,
    his: List[Dict[str, str]] = [],
    max_length: int = 4096,
    top_p: float = 0.95,
    temperature: float = 0.1,
    **kwargs,
):
    messages = [
        {"role": message["role"], "content": process_input(message["content"])}
        for message in his
    ] + [{"role": "user", "content": process_input(ins)}]

    text = self.processor.apply_chat_template(messages, add_generation_prompt=True, tokenize=False)
    audios = []
    for message in messages:
        if isinstance(message["content"], list):
            for ele in message["content"]:
                if ele["type"] == "audio":
                    audio_data = ele["audio_url"].split(",")[1]
                    audio_bytes = base64.b64decode(audio_data)
                    audio = librosa.load(io.BytesIO(audio_bytes), sr=self.processor.feature_extractor.sampling_rate)[0]
                    audios.append(audio)
        
    inputs = self.processor(text=text, audios=audios, return_tensors="pt", padding=True)
    inputs.input_ids = inputs.input_ids.to(self.device)

    start_time = time.monotonic()
    generate_ids = self.generate(**inputs, max_length=max_length)
    generate_ids = generate_ids[:, inputs.input_ids.size(1):]

    response = self.processor.batch_decode(generate_ids, skip_special_tokens=True, clean_up_tokenization_spaces=False)[0]
    time_cost = time.monotonic() - start_time

    return [
        (
            response,
            {
                "metadata": {
                    "request_id": "",
                    "input_tokens_count": inputs.input_ids.size(1),
                    "generated_tokens_count": generate_ids.size(1),
                    "time_cost": time_cost,
                    "first_token_time": 0,
                    "speed": float(generate_ids.size(1)) / time_cost if time_cost > 0 else 0,
                }
            },
        )
    ]

def init_model(model_dir, infer_params: Dict[str, str] = {}, sys_conf: Dict[str, str] = {}):
    processor = AutoProcessor.from_pretrained(model_dir)
    model = Qwen2AudioForConditionalGeneration.from_pretrained(model_dir, device_map="auto")
    model.processor = processor
    import types
    model.stream_chat = types.MethodType(stream_chat, model)
    model.get_meta = types.MethodType(get_meta, model)    
    return (model, None)


import json
import os

import argparse
import deepspeed
import deepspeed.comm as dist
import numpy as np
import sentencepiece as spm
import torch

from .base_model.configuration_baichuan import BaiChuanConfig
from .base_model.modeling_baichuan import BaiChuanForCausalLM


def get_argument_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", type=str, default="data_dir",
                        help="Text files to do pre-train on")

    parser.add_argument("--tokenizer_path", type=str,
                        default="tokenizer.model",
                        help="Tokenizer model file path")

    parser.add_argument("--max_length", type=int, default=4096,
                        help="Max tokens per sentence in corpus")

    parser.add_argument("--steps_per_epoch", type=int, default=4096,
                        help="Step intervals to save checkpoint")

    parser.add_argument("--checkpoint_saving_path", type=str,
                        default="checkpoints",
                        help="Path to store checkpoint files")

    parser.add_argument("--local_rank", type=int, default=-1,
                        help="Reserved for deepspeed framework")
    return parser


arg_parser = get_argument_parser()
arg_parser = deepspeed.add_config_arguments(arg_parser)
args = arg_parser.parse_args()
deepspeed.init_distributed()


class DataEngine():
    def __init__(self, data_dir, tokenizer_path, micro_batch_size, max_length):
        self.MIN_TEXT_LEN = 20
        self.EOS_TOKEN_ID = 2
        self.data_dir = data_dir
        self.sp = spm.SentencePieceProcessor()
        self.sp.Load(tokenizer_path)
        self.micro_batch_size = micro_batch_size
        self.max_length = max_length
        self.data = []
        self.global_input_paths = [self.data_dir + "/" + x
                                   for x in os.listdir(self.data_dir)]
        self.local_input_paths = [x for i, x in
                                  enumerate(self.global_input_paths)
                                  if i % dist.get_world_size() == dist.get_rank()]

    def load_data(self):
        for file_path in self.local_input_paths:
            data = []
            with open(file_path, encoding="utf-8", errors="ignore") as f:
                for line_id, line in enumerate(f):
                    line_json = json.loads(line.strip())
                    text = line_json("instruction") + line_json("input") + line_json("output")
                    cc = self.sp.EncodeAsIds(text) + [self.EOS_TOKEN_ID]
                    if len(cc) < self.MIN_TEXT_LEN:
                        cc = []
                    data.extend(cc)
                    if len(data) >= self.micro_batch_size * (self.max_length + 1):
                        index = self.micro_batch_size * (self.max_length + 1)
                        self.data.append(data[:index])
                        data = []
        return

    def get_data(self):
        data = self.data.pop(0)
        seq = np.asarray(data).reshape(self.micro_batch_size, self.max_length + 1)
        data = torch.LongTensor(seq)
        data = data.cuda(non_blocking=True)
        return data


def prepare_data():
    data_dir = args.data_dir
    tokenizer_path = args.tokenizer_path
    ds_config = json.load(open(args.deepspeed_config))
    micro_batch_size = ds_config["train_micro_batch_size_per_gpu"]
    max_length = args.max_length
    data_engine = DataEngine(data_dir, tokenizer_path, micro_batch_size, max_length)
    data_engine.load_data()
    return data_engine


def prepare_model():
    with deepspeed.zero.Init(config_dict_or_path=args.deepspeed_config,
                             enabled=True,
                             mem_efficient_linear=False,
                             mpu=None):
        model = BaiChuanForCausalLM(BaiChuanConfig())

    model_parameters = filter(lambda p: p.requires_grad, model.parameters())
    model_engine, _, _, _ = deepspeed.initialize(args=args,
                                                 model=model,
                                                 optimizer=None,
                                                 model_parameters=model_parameters)
    return model_engine


def train(data_engine, model_engine):
    model_engine.train()
    step = 0
    while step < args.steps_per_epoch:
        data = data_engine.get_data()
        loss = model_engine(data, labels=data).loss
        model_engine.backward(loss)
        model_engine.step()
        step += 1
    return


if __name__ == "__main__":
    data_engine = prepare_data()
    model_engine = prepare_model()
    epoch = 0
    while True:
        train(data_engine, model_engine)
        epoch += 1
        model_engine.save_checkpoint(f"{args.checkpoint_saving_path}",
                                     tag=f"Epoch-{epoch}")
